#ifndef A_LEAK_DETECTABLE_H
#define A_LEAK_DETECTABLE_H

#include "ArduinoSTL.h"
#include "PulseCounter.h"
#include "LeakDetectionConfig.h"
#include "LimitsDefinition.h"
#include "DateTime.h"
#include "Pins.h"

//====================web server========================

// if the webserver is enabled,
// include these files as well
#ifdef WEB_SERVER
# include <map>
# include <HTMLDataSource.h>
#endif

//====================debugging=========================

// if debugging is enabled,
// include this file as well
#ifdef DEBUG
# include <ILoggable.h>
#endif

//================email notifications===================

// if email-notifications are 
// enabled, include this file 
// as well
#ifdef EMAIL_NOTIFICATION
# include <EmailSender.h>
#endif

//==================LCD display=========================

// if the LCD display is enabled,
// include this file as well
#ifdef LCD_DISPLAY
# include <IDisplayable.h>
#endif

//==============ALeakDetectable class===================

/// \author silhavyj A17B0362P
///
/// This class sepresents an abstraction of a water leak
/// detection algorithm. It holds all the common variables,
/// defines the common methods, etc.
class ALeakDetectable {

//====================public data=======================
public:
    /// The type of the water leak detection algorithm
    enum Type {
        Low,    ///< low water leak detection algorithm
        High,   ///< high water leak detection algorithm
        Total   ///< total water leak detection algorithm
    };

//===================protected data=====================
protected:
    PulseCounter *pulseCounter;         ///< instance of a pulse counter (input of the system)
    LeakDetectionConfig_t normalConfig; ///< normal configuarion for when the house is occupied 
    LeakDetectionConfig_t alarmConfig;  ///< alarm configuration for when there's nobody in the house
    LeakDetectionConfig_t *config;      ///< pointer to the current configuration
    int bypass;                         ///< bypass pin
    int oldBypassValue;                 ///< old bypass value (used for sending e-mails)
    int active;                         ///< indication if a leak has been detected + bypass is off
    int stateOfHomeAlarm;               ///< the current state of the home alarm
    bool detected;                      ///< indication if a leak has been detected
    int resetWhenDetected;              ///< the reset value (%) when a leak was detected
    String timeOfDetection;             ///< time when the leak has been detected
    Type type;                          ///< the type of the water leak detection algorithm

    LeakDetectionConfig_t newNormalConfig; ///< new parameters of the normal config when the user changes them
    LeakDetectionConfig_t newAlarmConfig;  ///< new parameters of the alarm config when the user changes them
    int requestUpdateNormalConfig;         ///< flag if the new normal config should be applied
    int requestUpdateAlarmConfig;          ///< flag if the new alarms config should be applied 

//==================private methods=====================
private:
    /// Test is a water leak has been detected.
    /// The implementation depends on the particular 
    /// algorithm.
    virtual void testActiveLeak() = 0;

    /// Test is a water leak has been detected.
    /// The implementation depends on the particular
    /// algorithm.
    virtual void testResetLeak() = 0;

//=================protected methods====================
protected:
    /// Applies the new config configuration
    ///
    /// This method is called on each reset of the algorithm.
    /// If the appropriate flag is set, the new settings will
    /// be put in place.
    void applyNewConfig();

    /// Returns the type of the water leak detection algorithm.
    /// \return the type of the detection algorithm
    String getType() const;

//==================public methods======================
public:
    /// Constructor of the class
    ///
    /// \param pulseCounter - instance of PulseCounter which is used
    ///                       as the main input of all water leak
    ///                       detection algorithms
    /// \param type - type of the water leak detection algorithm (high, low, total)
    /// \param normalConfig - default normal configuration (defined in LimitsDefinition.h)
    /// \param alarmConfig - default alarms configuration (defined in LimitsDefinition.h)
    ALeakDetectable(PulseCounter *pulseCounter,
                    Type type,
                    LeakDetectionConfig_t normalConfig,
                    LeakDetectionConfig_t alarmConfig);

    /// Destructor of the class
    virtual ~ALeakDetectable() {};

    /// Returns a flag whether of not a leak has been detected and the bypass is off
    /// \return flag (1/0) indicating to close up the main valve
    int isActive() const { return active; }

    /// Changes the state of the home alarm
    /// \param state new state of the home alarm
    void changeStateOfHomeAlarm(bool state);

    /// Returns the current alarm config
    /// \return the current alarm config
    LeakDetectionConfig_t getAlarmConfig();

    /// Returns the current normal config
    /// \return the current normal config
    LeakDetectionConfig_t getNormalConfig();

    /// Returns a string containing the current settings of the detection algorithm
    ///
    /// It contains both kinds of configurations. The settings will be stored
    /// on the SD card so it can be loaded on the next start of the system.
    ///
    /// \return settings in a string format
    virtual String getFormatOfSettingsToSave() = 0;

    /// Updates the normal settings
    ///
    /// The settings won't be applied right away. Instead, the flag will be set
    /// and the new setting will be put in place when the algorithm resets.
    ///
    /// \param newNormalConfig the new normal settings
    void updateNormalConfig(LeakDetectionConfig_t newNormalConfig);

    /// Updates the alarm settings
    ///
    /// The settings won't be applied right away. Instead, the flag will be set
    /// and the new settings will be put in place when the algorithm resets.
    ///
    /// \param newAlarmConfig the new alarm settings
    void updateAlarmConfig(LeakDetectionConfig_t newAlarmConfig);

    /// This method is called for updating its state.
    ///
    /// It updates the algorithm in order to find out if there's a leak
    /// taking place.
    virtual void update() = 0;

    /// This method resets the algorithm
    ///
    /// This is called either when the user manually presses the
    /// reset button or when the algorithm resets by itself.
    virtual void reset() = 0;

//===============email notifications====================
#ifdef EMAIL_NOTIFICATION
    /// Sends off an email regarding the settings of the algorithm
    ///
    /// This method sends two kinds of emails. The first type is when
    /// the settings have been changed by the user but they have not been
    /// put in place yet. The second type is when the settings are 
    /// applied on the reset of the algorithm.
    ///
    /// \param emailType type of the email
    /// \param oldSettings old configuration
    /// \param newSettings new configuration
    /// \param alarm type of the settings (alarm/normal)
    void sendEmailAboutSettings(EmailSender::Type emailType,
                                LeakDetectionConfig_t& oldSettings,
                                LeakDetectionConfig_t& newSettings,
                                bool alarm) const;
#endif
};

#endif