#ifndef UNTITLED_DAILY_OVERVIEW_H
#define UNTITLED_DAILY_OVERVIEW_H

#include "Arduino.h"
#include "IControllable.h"
#include "Setup.h"
#include "DateTime.h"
#include "HighLeakDetection.h"
#include "LowLeakDetection.h"
#include "TotalLeakDetection.h"
#include "Consumption.h"
#include "EmailSender.h"
#include "LimitsDefinition.h"

//===============DailyOverview class====================

/// \author silhavyj A17B0362P
///
/// This class collects information to be sent to the user
/// as a daily overview. Such information includes daily
/// water consumption, how close each type of water leak
/// was from being detected [%], etc.
class DailyOverview : public IControllable
#ifdef LCD_DISPLAY
 , public IDisplayable
#endif
{
//====================private data======================
private:
    HighLeakDetection *highLeakDetection;    ///< instance of a high-water leak detection algorithm
    LowLeakDetection *lowLeakDetection;      ///< instance of a low-water leak detection algorithm
    TotalLeakDetection *totalLeakDetection;  ///< instance of a total-water leak detection algorithm

    Consumption *dailyConsumption;    ///< instance of Consumption (daily water consumption)
    Consumption *monthlyConsumption;  ///< instance of Consumption (monthly water consumption)
    RTCDateTime initialDateTime;      ///< instance of RTCDateTime (real-time module)

    float maxHighLeakPercentage;   ///< max [%] high-water leak detection within the monitoring period (extreme)
    float maxLowLeakPercentage;    ///< max [%] low-water leak detection within the monitoring period (extreme)
    float maxTotalLeakPercentage;  ///< max [%] total-water leak detection within the monitoring period (extreme)

//==================private methods=====================
private:
    /// Resets all the variables at the end of the day
    void reset();

//==================public methods======================
public:
    /// Constructor of the class
    ///
    /// It takes already existing instances as parameters
    /// so it can gather information off them
    ///
    /// \param highLeakDetection instance of HighLeakDetection
    /// \param lowLeakDetection instance of LowLeakDetection
    /// \param totalLeakDetection instance of LowLeakDetection
    /// \param dailyConsumption instance of Consumption (daily)
    /// \param monthlyConsumption instance of Consumption (monthly)
    DailyOverview(HighLeakDetection *highLeakDetection,
                  LowLeakDetection *lowLeakDetection,
                  TotalLeakDetection *totalLeakDetection,
                  Consumption *dailyConsumption,
                  Consumption *monthlyConsumption);

    /// Updates the class
    /// 
    /// It updates the class along with all the information it is
    /// collecting. Also, using the RTCDateTime, it resets all the
    /// variables at the end of the day.
    void update() override;

//===================LCD display========================
#ifdef LCD_DISPLAY
    /// Returns the content of the row given as a parameter.
    ///
    /// This class represents one page on the LCD display.
    /// Each of the 4 rows has different content identified
    /// by the number of the row.
    ///
    /// \param row number of the row
    /// \return content to be displayed on that row of the LCD
    const String getRow(int row) const override;
#endif
};

#endif