#ifndef UNTITLED_DATETIME_H
#define UNTITLED_DATETIME_H

#include "Wire.h"
#include "DS3231.h"
#include "Arduino.h"

#include "Setup.h"

// if the webserver is enabled,
// include these files as well
#ifdef WEB_SERVER
# include <map>
# include <HTMLDataSource.h>
#endif

/// Referring to the data type unsigned long
/// as time_t. It is used for variables
/// holding time information in milliseconds.
typedef unsigned long time_t;

/// Calculates the time difference between
/// two times given as parameters.
///
/// Since the data type of both parameters is
/// unsigned long, it is not necessary to check
/// if the result of the operation is negative
/// or not. It is taken care of all automatically
/// by the data type itself. Both parameters are in
/// milliseconds.
///
/// @param time1 time x
/// @param time2 time y
/// @returns the result of the operation (x-y)[ms]
#define DELTA_TIME(time1, time2) \
((time_t)((time1) - (time2)))

//==============TIME CONSTANTS DEFINITIONS==============

/// number of milliseconds in onc second
#define MILLIS_PER_SEC 1000UL
/// number of seconds in one minute
#define SECS_PER_MIN   60UL
/// number of minutes in one hour
#define MINS_PER_HOUR  60UL
/// numer of hours in one day
#define HOURS_PER_DAY  24UL

/// number of milliseconds in one minute
#define MILLIS_PER_MIN (MILLIS_PER_SEC * SECS_PER_MIN)
/// number of milliseconds in one hour
#define MILLIS_PER_HOD (MILLIS_PER_MIN * MINS_PER_HOUR)
/// number of milliseconds in one day
#define MILLIS_PER_DAY (MILLIS_PER_HOD * HOURS_PER_DAY)

//====================TIME CONVERSION===================

/// Converts seconds given as a parameter to milliseconds.
/// @param s number of seconds
/// @returns time given as a parameter [s] in milliseconds
#define SEC_TO_MILLIS(s) ((time_t)(s) * MILLIS_PER_SEC)

/// Converts minutes given as a parameter to milliseconds.
/// @param m number of minutes
/// @returns time given as a parameter [min] in milliseconds
#define MIN_TO_MILLIS(m) ((time_t)(m) * MILLIS_PER_MIN)

/// Converts hours given as a parameter to milliseconds.
/// @param h number of hours
/// @returns time given as a parameter [hrs] in milliseconds
#define HOD_TO_MILLIS(h) ((time_t)(h) * MILLIS_PER_HOD)

/// Converts days given as a parameter to milliseconds.
/// @param d number of days
/// @returns time given as a parameter [days] in milliseconds
#define DAY_TO_MILLIS(d) ((time_t)(d) * MILLIS_PER_DAY)

//====================TIME FORMATTING===================

/// Calculates the number of milliseconds when
/// formatting the time given in milliseconds.
/// @param time in milliseconds
/// @returns number of milliseconds when converting
///         into format [d:h:m:s:ms]
#define MILLIS(time) ((time) % MILLIS_PER_SEC)

/// Calculates the number of seconds when
/// formatting the time given in milliseconds.
/// @param time in milliseconds
/// @returns number of seconds when converting
///         into format [d:h:m:s:ms]
#define SECS(time) (((time) / MILLIS_PER_SEC) % SECS_PER_MIN)

/// Calculates the number of minutes when
/// formatting the time given in milliseconds.
/// @param time in milliseconds
/// @returns number of minutes when converting
///         into format [d:h:m:s:ms]
#define MINS(time) (((time) / MILLIS_PER_MIN) % MINS_PER_HOUR)

/// Calculates the number of hours when
/// formatting the time given in milliseconds.
/// @param time in milliseconds
/// @returns number of hours when converting
///         into format [d:h:m:s:ms]
#define HOURS(time) (((time) / MILLIS_PER_HOD) % HOURS_PER_DAY)

/// Calculates the number of days when
/// formatting the time given in milliseconds.
/// @param time in milliseconds
/// @returns number of days when converting
///         into format [d:h:m:s:ms]
#define DAYS(time) ((time) / MILLIS_PER_DAY)

/// Converts the number given as a parameter
/// into a 2 digit string.
///
/// For example, number 9 becomes "09".
///
/// @param x number that is going to be formatted
/// @returns formatted string - 2 digits
#define FORMAT_2_DIGITS(x) \
((x) >= 10 ? String(x) : ("0" + String(x)))

/// Converts the number given as a parameter
/// into a 3 digit string.
///
/// For example, number 9 becomes "009" or
/// number 15 becomes "015".
///
/// @param x number that is going to be formatted
/// @returns formatted string - 3 digits
#define FORMAT_3_DIGITS(x) \
((x) >= 100 ? String(x) : ( \
(x) >= 10 ? ("0" + String(x)) : ("00" + String(x))))

/// Converts a time in milliseconds into a formatted string.
/// The format of the final string is [days:hrs:mins:secs:millis].
/// @param time in millisecond that is going to be formatted
/// @returns formatted string as [d:h:m:s:ms]
#define FORMAT_TIME(time) \
(String(DAYS(time)) + ":" + \
 String(FORMAT_2_DIGITS(HOURS(time))) + ":" + \
 String(FORMAT_2_DIGITS(MINS(time))) + ":" + \
 String(FORMAT_2_DIGITS(SECS(time))) + ":" + \
 String(FORMAT_3_DIGITS(MILLIS(time))) \
)

//=================HTMLDataSource class=================

/// \author silhavyj A17B0362P
///
/// This class holds information about datetimes.
/// This includes values such as the system start time,
/// current datatime, etc.
/// If the webserver is enabled, this class also works
/// a source of data for the HTML content.
class DateTime
#ifdef WEB_SERVER
 : public HTMLDataSource
#endif
{
//====================private data======================
private:
    static DateTime* instance;  ///< the instance of the class (Singleton)
    DS3231 rtc;                 ///< real-time module
    String startDate;           ///< start day of the system
    String startTime;           ///< start time of the system

//==================private methods=====================
private:
    /// Constructor of the class
    DateTime();

    /// Destructor of the class
    ~DateTime();

    /// Copy constructor of the class
    DateTime(DateTime const&) {};

    /// Assignment operator of the class
    DateTime& operator=(DateTime const&) { return *this; }

//==================public methods======================
public:
    /// Returns the instance of the class
    ///
    /// If the instance has not been created, it will 
    /// create it and then return it.
    ///
    /// \return the instance of the class
    static DateTime* getInstance();

    /// Returns the current datatime in a string format
    /// \return the current datatime as a string
    String getDateTimeStr() const;

    /// Returns the current day in a string format
    /// \return the current day as a string
    String getDateStr() const;

    /// Returns the current time in a string format
    /// \return the current time as a string
    String getTimeStr() const;

    /// Returns the current datatime
    /// \return the current datetime
    RTCDateTime getDateTime() const;

//====================web server========================
#ifdef WEB_SERVER
private:
    /// A map of different values (keys) and their associated functions
    /// which returns the appropriate values.
    std::map<int, String (*)(const DateTime& dateTime)> htmlData;

public:
    /// Since this class is registered as a source of data for the
    /// HTML content, it needs return the appropriate value according 
    /// to the id given as a parameter.
    /// 
    /// \param id - the id of the piece of data
    /// \return If the class holds this piece of information, it will
    ///         return it. Otherwise, "UNDEFINED" will be returned.
    const String getHTMLData(const int id) const;

private:
    /// Associated function for returing the start day of the device
    /// \param dateTime - the instance of the DateTime class
    /// \return the device start day
    friend String HTML_startDate(const DateTime& dateTime);

    /// Associated function for returing the start time of the device
    /// \param dateTime - the instance of the DateTime class
    /// \return the device start time
    friend String HTML_startTime(const DateTime& dateTime);
#endif
};

#endif
