#ifndef UNTITLED_EMAIL_SENDER_H
#define UNTITLED_EMAIL_SENDER_H

#include "SPI.h"
#include "Ethernet.h"
#include "SD.h"
#include "map"
#include "Setup.h"

//====================web server========================

// if the webserver is enabled,
// include this file as well
#ifdef WEB_SERVER
# include <HTMLDataSource.h>
#endif

//================constant definitions==================

#define arduinoIP "192.168.1.123"                       // ip address of Arduino

#define SMTPServerAddress "mail.smtp2go.com"            // address of the smt2go server
#define SMTPServerPort 2525                             // port of the service

#define emailAddress ""  // Arduino's e-mail address
#define auth64Login  ""  // Arduino's e-mail login
#define auth64Pass   ""  // Arduino's e-mail password

//=================EmailSender class====================

/// \author silhavyj A17B0362P
///
/// This calass is designed as a Singleton and is
/// used for sending e-mails to the user.
class EmailSender
    #ifdef WEB_SERVER
        : public HTMLDataSource
    #endif
{
//====================public data=======================
public:
    /// different types of e-mail that
    /// can be sent to the user
    enum Type { 
        BOOTING,             ///< when the system boots up
        DAILY_OVERVIEW,      ///< daily overview about the system
        LEAK_DETECTED,       ///< when a leak has been detected
        VALVE_STATE,         ///< when the state of the valve has changed
        RESET,               ///< when the device resets
        ALARM,               ///< when the state of the home alarm changes
        CHANGED_SETTINGS,    ///< when the user changes settings
        APPLIED_NEW_SETTING, ///< when the new settings are applied
        BYPASS,              ///< when a bypass changes

        COUNT = 9            ///< total number of different kinds of e-mails
    };

//====================private data======================
private:
    std::map<Type, bool> enabled;   ///< map holding info about each time of e-mail - whether it's enabled or not 
    static EmailSender* instance;   ///< the instance of EmailSender
    EthernetClient client;          ///< EthernetClient used to connect to the smt2go server

    byte mac[6] = { 0x00, 0xAA, 0xBB, 0xCC, 0xDE, 0x02 }; ///< mac address of the ethernet shield
    String receiverEmailAddress = "jakub.sil@seznam.cz";  ///< the default user's email address 

//==================private methods=====================
private:
    /// Constructor of the class
    EmailSender();

    /// Descrutor of the class
    ~EmailSender();

    /// Copy constructor of the class
    EmailSender(EmailSender const&) {};

    /// Assignmet operator of the class
    EmailSender& operator=(EmailSender const&) { return *this; }

//------------------------------------------------------

    /// Initializes the ethernet shield
    ///
    /// This method initializes the the ethernet shield only if 
    /// it has not been initialized yet.
    ///
    /// \return 1 if the shield has been initialized successfully, 0 otherwise.
    int setEthernet();

    /// Reads data using the ethernet client comming from the smtp2go server
    /// \return 0 if the reading fails, otherwise 1
    byte eRcv();

    /// Closes the ethernet client after the connection has failed. 
    void efail();

    /// Sends an e-mail off to the smtp2go server.
    /// \param subject of the e-mail
    /// \param data content of the e-mail
    /// \return 1 if the e-mail has been sent off successfully, 0 otherwise.
    byte sendEmail(String subject, String data);

//==================public methods======================
public:
    /// Returns the instance of the class
    /// \return the instance of EmailSender
    static EmailSender* getInstance();

    /// Sends an e-mail to the smtp2go server
    /// \param type of the e-mail
    /// \param subject of the e-mail
    /// \param data content of the e-mail
    void sendEmail(Type type, String subject, String data);

    /// Enables/disables the particular type of e-mail notification
    /// \param type of e-mail notification
    /// \param state true/false - enable/disable
    void enableNotification(Type type, bool state);

    /// Changes the user's e-mail address
    /// \param email the new e-mail address all e-mails will be sent to
    void setReceiverEmailAddress(String email);

    /// Returns settings
    ///
    /// It returns the user's e-mail address as well as information
    /// about which e-mail notifications are enabled/disabled so they
    /// can be stored on the SD card.
    /// 
    /// \return settings in a string format
    String getFormatOfSettingsToSave();

//====================web server========================
#ifdef WEB_SERVER
private:
    /// A map of different values (keys) and their associated functions
    /// which returns the appropriate values.
    std::map<int, String (*)(const EmailSender& leaksController)> htmlData;

public:
    /// Since this class is registered as a source of data for the
    /// HTML content, it needs return the appropriate value according 
    /// to the id given as a parameter.
    /// 
    /// \param id - the id of the piece of data
    /// \return If the class holds this piece of information, it will
    ///         return it. Otherwise, "UNDEFINED" will be returned.
    const String getHTMLData(const int id) const;

private:
    /// Associated function for returing the state of the 'bootup' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_BOOTING(const EmailSender& emailSender);

    /// Associated function for returing the state of the 'daily overview' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_DAILY_OVERVIEW(const EmailSender& emailSender);

    /// Associated function for returing the state of the 'leak detected' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_LEAK_DETECTED(const EmailSender& emailSender);

    /// Associated function for returing the state of the 'valve state' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_VALVE_STATE(const EmailSender& emailSender);

    /// Associated function for returing the state of the 'system reset' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_RESET(const EmailSender& emailSender);

    /// Associated function for returing the state of the 'home alarm state' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_ALARM(const EmailSender& emailSender);

    /// Associated function for returing the state of the 'changed settings' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_CHANGED_SETTINGS(const EmailSender& emailSender);

    /// Associated function for returing the state of the 'applied settings' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_APPLIED_NEW_SETTING(const EmailSender& emailSender);

    /// Associated function for returing the state of the 'bypass state' notification
    /// \param emailSender - the instance of the EmailSender class
    /// \return "checked" if the notification is enabled, "" otherwise
    friend String HTML_Notification_BYPASS(const EmailSender& emailSender);

    /// Associated function for returing the user's e-mail address
    /// \param emailSender - the instance of the EmailSender class
    /// \return the receiver's e-mail address
    friend String HTML_ReceiverEmailAddress(const EmailSender& emailSender);
#endif
};

#endif