#ifndef UNTITLED_FREE_MEMORY_MEASUREMENT_H
#define UNTITLED_FREE_MEMORY_MEASUREMENT_H

#ifdef __arm__
    extern "C" char* sbrk(int incr);
#else  // __ARM__
    extern char *__brkval;
#endif  // __arm__

#include "Arduino.h"
#include "Setup.h"

//====================debugging=========================

// if debugging is enabled,
// include this file as well
#ifdef DEBUG
# include <ILoggable.h>
#endif

//====================web server========================

// if the webserver is enabled,
// include these files as well
#ifdef WEB_SERVER
# include <map>
# include <HTMLDataSource.h>
#endif

//==================LCD display=========================

// if the LCD display is enabled,
// include this file as well
#ifdef LCD_DISPLAY
# include <IDisplayable.h>
#endif

//===========FreeMemoryMeasurement class================

/// \author silhavyj A17B0362P
///
/// This class provides information about the current
/// use of the the SRAM memory. The particular value
/// depends on the moment the method is called as the
/// system might be performing different memory-consuming
/// tasks. This information is used as an indication what
/// the device is not running out of memory.
class FreeMemoryMeasurement
        #ifdef DEBUG
            : public ILoggable
        #endif
        #ifdef WEB_SERVER
        # ifdef DEBUG
            , public HTMLDataSource
        # else
            : public HTMLDataSource
        # endif
        #endif
        #ifdef LCD_DISPLAY
        #if defined(DEBUG) || defined(WEB_SERVER)
            , public IDisplayable
        # else
            : public IDisplayable
        # endif
        #endif
{
//====================private data======================
private:
    static FreeMemoryMeasurement* instance; ///< the instance of the class

//==================private methods=====================
private:
    /// Constructor of the class
    FreeMemoryMeasurement();

    /// Copy constructor of the class
    FreeMemoryMeasurement(FreeMemoryMeasurement const&) {};

    /// Assignment operator of the class
    FreeMemoryMeasurement& operator=(FreeMemoryMeasurement const&) { return *this; }

    /// Return the amount of RAM currently being used
    /// \return the amount of RAM being used at the moment
    int getFreeRAM() const;

//==================public methods======================
public:
    /// Returns the instance of the class
    ///
    /// If the instance has not been created, it will 
    /// create it and then return it.
    ///
    /// \return the instance of the class
    static FreeMemoryMeasurement* getInstance();

//====================debugging=========================
#ifdef DEBUG
    /// Returns the log description (body) of the class
    /// \return the content of the log being printed out
    const String getLogDescription() const;

    /// Returns the log id of the class
    /// \return the log id
    const String getLogID() const;
#endif

//==================LCD display=========================
#ifdef LCD_DISPLAY
    /// Returns the content of the row given as a parameter.
    ///
    /// This class represents one page on the LCD display.
    /// Each of the 4 rows has different content identified
    /// by the number of the row.
    ///
    /// \param row number of the row
    /// \return content to be displayed on that row of the LCD
    const String getRow(int row) const override;
#endif

//====================web server========================
#ifdef WEB_SERVER
private:
    /// A map of different values (keys) and their associated functions
    /// which returns the appropriate values.
    std::map<int, String (*)(const FreeMemoryMeasurement& freeMemoryMeasurement)> htmlData;

public:
    /// Since this class is registered as a source of data for the
    /// HTML content, it needs return the appropriate value according 
    /// to the id given as a parameter.
    /// 
    /// \param id - the id of the piece of data
    /// \return If the class holds this piece of information, it will
    ///         return it. Otherwise, "UNDEFINED" will be returned.
    const String getHTMLData(const int id) const override;

private:
    /// Associated function for returing amount of memory being used at the moment
    /// \param freeMemoryMeasurement - the instance of the FreeMemoryMeasurement class
    /// \return the memory usage [B]
    friend String HTML_freeRAM(const FreeMemoryMeasurement& freeMemoryMeasurement);
#endif
};

#endif