#ifndef HIGH_LEAK_DETECTION_H
#define HIGH_LEAK_DETECTION_H

#include "Setup.h"
#include "ALeakDetectable.h"

//=============HighLeakDetection class==================

/// \author silhavyj A17B0362P
///
/// This class implements the high-water leak detection
/// algorithm and everything that is associated with it.
class HighLeakDetection : public ALeakDetectable
        #ifdef DEBUG
         , public ILoggable
        #endif
        #ifdef WEB_SERVER
         , public HTMLDataSource
        #endif
        #ifdef LCD_DISPLAY
         , public IDisplayable
        #endif
{
//===================protected data=====================
protected:
    int pulseCount; ///< number of pulses detected so far

//=================protected methods====================
protected:
    /// Tests if a high-water leak has occurred, and
    /// if so, it will set the appropriate variables and flags.
    void testActiveLeak() override;

    /// Tests if the high-water the algorithm should be reset
    /// based on inactivity of pulse on the input pin
    void testResetLeak() override;

//==================public methods======================
public:
    /// Constructor of the class
    ///
    /// \param pulseCounter instance of PulseCounter (input)
    /// \param type of the algorithm (high-water)
    /// \param normalConfig default normal settings
    /// \param alarmConfig default alarm settings
    HighLeakDetection(PulseCounter *pulseCounter,
                      Type type,
                      LeakDetectionConfig_t normalConfig,
                      LeakDetectionConfig_t alarmConfig);

    /// Descrutor of the class
    ~HighLeakDetection() {}

    /// Updates the algorithm
    void update() override;

    /// Resets the algorithm
    void reset() override;

    /// Returns the current settings of the algorithm
    ///
    /// This method is called when storing data on the SD card
    /// so they can be loaded when the system reboots
    ///
    /// \return the current settings in a string format 
    String getFormatOfSettingsToSave() override;

    /// Returns percentage information about how close a high-water leak is from being detected
    ///
    /// The particular value is worked out based on the current settings.
    /// This piece of information is then displayed to the user.
    ///
    /// \return information about how close a high-water leak is from being detected
    float getPercentLeakDetectionReset() const;

    /// Returns percentage information about how close the algorithm is from being reset
    ///
    /// The particular value is worked out based on the current settings.
    /// This piece of information is then displayed to the user.
    ///
    /// \return information about how close the algorithm is from being reset
    float getPercentLeakDetected() const;

//====================debugging=========================
#ifdef DEBUG
public:
    /// Returns the log description (body) of the class
    /// \return the content of the log being printed out
    const String getLogDescription() const override;

    /// Returns the log id of the class
    /// \return the log id
    const String getLogID() const override;
#endif

//==================LCD display=========================
#ifdef LCD_DISPLAY
    /// Returns the content of the row given as a parameter.
    ///
    /// This class represents one page on the LCD display.
    /// Each of the 4 rows has different content identified
    /// by the number of the row.
    ///
    /// \param row number of the row
    /// \return content to be displayed on that row of the LCD
    const String getRow(int row) const override;
#endif

//====================web server========================
#ifdef WEB_SERVER
protected:
    /// A map of different values (keys) and their associated functions
    /// which returns the appropriate values.
    std::map<int, String (*)(const HighLeakDetection& highLeakDetection)> htmlData;

public:
    /// Since this class is registered as a source of data for the
    /// HTML content, it needs return the appropriate value according 
    /// to the id given as a parameter.
    /// 
    /// \param id - the id of the piece of data
    /// \return If the class holds this piece of information, it will
    ///         return it. Otherwise, "UNDEFINED" will be returned.
    const String getHTMLData(const int id) const override;

private:
    /// Associated function for returing information about the state of the bypass
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return "OFF"/"ON" according to the current state of the bypass
    friend String HTML_bypass(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing information about a high-water leak being detected (1)
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return "ON" if a high-water leak has occurred, "OFF" otherwise
    friend String HTML_detected(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing information about a high-water leak being detected (2)
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return "Detected" if a high-water leak has occurred, "Undetected" otherwise
    friend String HTML_detectedText(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the detection time of a high-water leak
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return the time of detection in a string format
    friend String HTML_detectedTime(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the [%] information about how close a leak is from being detected
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return [%] information about how close a leak is from being detected
    friend String HTML_activePercentage(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing a color according to [%] information about how close a leak is from being detected
    ///
    /// For instance,  0 -  59% -> gray
    ///               60 -  79% -> yellow
    ///               80 -  89% -> orange    
    ///               90 - 100% ->red
    ///
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return color according to the state of the algorithm - detected [%]
    friend String HTML_activePercentageColor(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the [%] information about how close the algorithm is from being reset
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return [%] information about how close the algorithm is from being reset
    friend String HTML_resetPercentage(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the limit (detection) in terms of the normal configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return the limit (detection) in the normal configuration
    friend String HTML_detectionLimit(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the reset time (days) in the normal configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return reset time (days) in the normal configuration
    friend String HTML_resetLimitDays(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the reset time (hours) in the normal configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return reset time (hours) in the normal configuration
    friend String HTML_resetLimitHours(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the reset time (mins) in the normal configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return reset time (mins) in the normal configuration
    friend String HTML_resetLimitMins(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the reset time (secs) in the normal configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return reset time (secs) in the normal configuration
    friend String HTML_resetLimitSecs(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the limit (detection) in terms of the alarm configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return the limit (detection) in the alarm configuration
    friend String HTML_detectionLimitAlarm(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the reset time (days) in the alarm configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return reset time (days) in the alarm configuration
    friend String HTML_resetLimitDaysAlarm(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the reset time (hours) in the alarm configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return reset time (hours) in the alarm configuration
    friend String HTML_resetLimitHoursAlarm(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the reset time (mins) in the alarm configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return reset time (mins) in the alarm configuration
    friend String HTML_resetLimitMinsAlarm(const HighLeakDetection& highLeakDetection);

    /// Associated function for returing the reset time (secs) in the alarm configuration
    /// \param highLeakDetection - the instance of the HighLeakDetection class
    /// \return reset time (secs) in the alarm configuration
    friend String HTML_resetLimitSecsAlarm(const HighLeakDetection& highLeakDetection);
#endif

//===============email notifications====================
#ifdef EMAIL_NOTIFICATION
private:
    /// Sends an e-mail when a high-water leak has been detected
    ///
    /// The message contains information such as the time of
    /// detection, state of the bypass button, and the limit
    /// that has been reached resulting in a high-water leak.
    virtual void sendEmailLeakDetected();
#endif
};

#endif