#ifndef LCD_CONTROLLER_H
#define LCD_CONTROLLER_H

#include "Arduino.h"
#include "Wire.h"
#include "LiquidCrystal_I2C.h"
#include "vector"
#include "Setup.h"
#include "DateTime.h"
#include "TimePeriod.h"
#include "IControllable.h"
#include "IDisplayable.h"

//====================debugging=========================

// if debugging is enabled,
// include this file as well
#ifdef DEBUG
# include <ILoggable.h>
#endif

#define LCD_ROWS 4 ///< number of rows on the LCD display

//===============LCDController class====================

/// \author silhavyj A17B0362P
///
/// This class works as a controller for the LCD display.
/// It holds a collection of classes capable of displaying
/// information on the display. Each class gets repeatedly
/// to display its content for a period of 3s.
class LCDController : public IControllable
#ifdef DEBUG
 , public ILoggable
#endif
{
//====================private data======================
private:
    static LCDController* instance;         ///< the instance of the class
    LiquidCrystal_I2C lcd = {0x27, 20, 4};  ///< 0x27 - address of the [20x4] LCD on I2C
    std::vector<IDisplayable *> items;      ///< collection of classes implementing the interface IDisplayable
    TimePeriod delay = SEC_TO_MILLIS(3);    ///< delay between displaying two pages
    int index;                              ///< index of the current page

//==================private methods=====================
private:
    /// Constructor of the class
    LCDController();

    /// Copy constructor of the class
    LCDController(LCDController const&) {};

    /// Assignment operator of the class
    LCDController& operator=(LCDController const&) { return *this; }

//==================public methods======================
public:
    /// Returns the instance of the class
    ///
    /// If the instance has not been created, it will 
    /// create it and then return it.
    ///
    /// \return the instance of the class
    static LCDController* getInstance();

    /// Updates the state of the class
    ///
    /// If the period of 3s has passed, it will display
    /// the content of the next page.
    void update() override;

    /// Adds another source of data for the LCD display (another class)
    /// \param item class to be added as another page
    void addItem(IDisplayable *item);

//====================debugging=========================
#ifdef DEBUG
public:
    /// Returns the log description (body) of the class
    /// \return the content of the log being printed out
    const String getLogDescription() const;

    /// Returns the log id of the class
    /// \return the log id
    const String getLogID() const;
#endif
};

#endif