#ifndef LEAKS_CONTROLLER_H
#define LEAKS_CONTROLLER_H

#include "Arduino.h"
#include "map"

#include "Setup.h"
#include "Pins.h"
#include "Button.h"
#include "IControllable.h"
#include "ALeakDetectable.h"
#include "PulseCounter.h"
#include "LimitsDefinition.h"

//====================debugging=========================

// if debugging is enabled,
// include this file as well
#ifdef DEBUG
# include <ILoggable.h>
#endif

//====================web server========================

// if the webserver is enabled,
// include these files as well
#ifdef WEB_SERVER
# include <HTMLDataSource.h>
# include <Consumption.h>
#endif

//================email notifications===================

// if email-notifications are 
// enabled, include this file 
// as well
#ifdef EMAIL_NOTIFICATION
# include <EmailSender.h>
#endif

//==================LCD display=========================

// if the LCD display is enabled,
// include this file as well
#ifdef LCD_DISPLAY
# include <IDisplayable.h>
#endif

//==============ALeakDetectable class===================

/// \author silhavyj A17B0362P
///
/// This class sepresents a controller for the water leak
/// detection part of the system. It takes care of reading
/// the input values, such as the state of the home alarm,
/// as well as upddating all the algorithms.
class LeaksController : public IControllable
#ifdef DEBUG
 , public ILoggable
#endif
#ifdef WEB_SERVER
 , public HTMLDataSource
#endif
#ifdef LCD_DISPLAY
 , public IDisplayable
#endif
{
//====================private data======================
private:
    static LeaksController* instance;                        ///< the instance of the class
    std::map<ALeakDetectable::Type, ALeakDetectable*> leaks; ///< collection (map) of water leak detection algorithms
    Button manualResetButton = RESET_PIN;                    ///< instance of Button (the reset buttton)

    int valveState;                    ///< state of the main valve
    int stateOfHomeAlarm;              ///< state of the home alarm
    int manualValveClose;              ///< state of the "manual close" button
    PulseCounter *pulseCounter = NULL; ///< instance of PulseCounter (output from the sensor)

//==================private methods=====================
private:
    /// Constructor of the class
    LeaksController();

    /// Copy constructor of the class
    LeaksController(LeaksController const&) {};

    /// Assignment operator of the class
    LeaksController& operator=(LeaksController const&) { return *this; }

//==================public methods======================
public:
    /// Returns the instance of the class
    ///
    /// If the instance has not been created, it will 
    /// create it and then return it.
    ///
    /// \return the instance of the class
    static LeaksController* getInstance();

    /// Adds another water leak detection algorithm to the collection
    ///
    /// This method is used in main.cpp during initialization
    ///
    /// \param type of the water leak detection algorithm
    /// \param leak pointer to an instance of ALeakDetectable (water leak detection algorithm)
    void addLeakDetection(ALeakDetectable::Type type, ALeakDetectable* leak);

    /// Return the number of water leak detection algorithm
    ///
    /// This is used when parsin an HTTP request so we know how many
    /// values there are supposed to be (settings).
    ///
    /// \return number of water leak detection algorithm in the collection
    int getNumberOfLeakDetections() const;

    /// Changes settings of a water leak detection algorithm
    /// \param type of the water leak detection algorithm
    /// \param detected limit (action) of a leak being detected
    /// \param reset reset time value of the algorithm
    /// \param alarm type of the settings (normal/alarm)
    void changeSettings(ALeakDetectable::Type type, unsigned long detected, unsigned long reset, bool alarm);

    /// Sets an instance of PulseCounter
    ///
    /// This method is used in main.cpp during initialization
    ///
    /// \param pulseCounter pointer to an instance of PulseCounter
    void setPulseCounter(PulseCounter *pulseCounter);

    /// Updates the class
    /// 
    /// It updates step by step all the variables and instance it holds.
    /// This process includes reading the state of the input pins, updating
    /// all water leak detection algorithms, and if necessary, closing the
    /// the main valve.
    void update() override;

    /// Returns all settings
    ///
    /// This method is used when storing data on the SD card. It returns
    /// the settings of all water leak detection algorithms.
    ///
    /// \return settings in a string format
    String getFormatOfSettingsToSave();

//==================UNIT testing========================
#ifdef UNIT_TEST
    /// Returns the state of the main valve
    ///
    /// This method is only used for unit testing. Other
    /// than that, the method is always disabled.
    ///
    /// \return the current state of the main valve
    int getValveState() const { return valveState; }
#endif

//====================debugging=========================
#ifdef DEBUG
public:
    /// Returns the log description (body) of the class
    /// \return the content of the log being printed out
    const String getLogDescription() const;

    /// Returns the log id of the class
    /// \return the log id
    const String getLogID() const;
#endif

//==================LCD display=========================
#ifdef LCD_DISPLAY
    /// Returns the content of the row given as a parameter.
    ///
    /// This class represents one page on the LCD display.
    /// Each of the 4 rows has different content identified
    /// by the number of the row.
    ///
    /// \param row number of the row
    /// \return content to be displayed on that row of the LCD
    const String getRow(int row) const override;
#endif

//====================web server========================
#ifdef WEB_SERVER
private:
    Consumption *dailyConsumptionCounter   = NULL; ///< daily water consumption
    Consumption *monthlyConsumptionCounter = NULL; ///< monthly water consumption

    /// A map of different values (keys) and their associated functions
    /// which returns the appropriate values.
    std::map<int, String (*)(const LeaksController& leaksController)> htmlData;

public:
    /// Since this class is registered as a source of data for the
    /// HTML content, it needs return the appropriate value according 
    /// to the id given as a parameter.
    /// 
    /// \param id - the id of the piece of data
    /// \return If the class holds this piece of information, it will
    ///         return it. Otherwise, "UNDEFINED" will be returned.
    const String getHTMLData(const int id) const override;

    /// Sets an instance of Consumption (daily water consumption)
    /// \param dailyConsumptionCounter pointer to an instance of Consumption
    void setDailyConsumptionCounter(Consumption *dailyConsumptionCounter);

    /// Sets an instance of Consumption (monthly water consumption)
    /// \param monthlyConsumptionCounter pointer to an instance of Consumption
    void setMonthlyConsumptionCounter(Consumption *monthlyConsumptionCounter);

private:
    /// Associated function for returing the current state of the main valve
    /// \param leaksController - the instance of the LeaksController class
    /// \return "ON"/"OFF" according to the state of the main valve
    friend String HTML_valve(const LeaksController& leaksController);

    /// Associated function for returing the current state of the home alarm
    /// \param leaksController - the instance of the LeaksController class
    /// \return "ON"/"OFF" according to the state of the home alarm
    friend String HTML_alarm(const LeaksController& leaksController);

    /// Associated function for returing the current state of the "manual close" button
    /// \param leaksController - the instance of the LeaksController class
    /// \return "ON"/"OFF" according to the state of the "manual close" button
    friend String HTML_manualClose(const LeaksController& leaksController);

    /// Associated function for returing the daily water consumption in liters
    /// \param leaksController - the instance of the LeaksController class
    /// \return daily water consumption in liters
    friend String HTML_dailyConsumption(const LeaksController& leaksController);

    /// Associated function for returing the monthly water consumption in liters
    /// \param leaksController - the instance of the LeaksController class
    /// \return monthly water consumption in liters
    friend String HTML_monthlyConsumption(const LeaksController& leaksController);
#endif
};

#endif