#ifndef PULSE_COUNTER_H
#define PULSE_COUNTER_H

#include "Arduino.h"
#include "Setup.h"
#include "DateTime.h"

//====================web server========================

// if the webserver is enabled,
// include these files as well
#ifdef WEB_SERVER
# include <map>
# include <HTMLDataSource.h>
#endif

//=================PulseCounter class===================

/// \author silhavyj A17B0362P
///
/// This class represents the main input of the system.
/// It keeps counting the pulses occuring on the input 
/// pin. A pulse is understood as the level of voltage
/// going from high to low.
class PulseCounter
#ifdef WEB_SERVER
 : public HTMLDataSource
#endif
{
//====================private data======================
private:
    int inputPin;           ///< number of the input pin
    time_t lastPulseTime;   ///< time when the last pulse was detected
    int previousInputValue; ///< previous state of the input pin
    int pulseDown;          ///< flag if a pulse just occurred

//==================public methods======================
public:
    /// Constructor of the class
    /// \param inputPin number of the input pin the flow sensor is connected to
    PulseCounter(int inputPin);

    /// Updates the class
    ///
    /// It reads the current state of the input pin
    /// and if it has gone from high to low, the flag
    /// will be set. This method is supposed to be
    /// called as frequently as possible so every
    /// pulse occurred will be detected.
    void update();

    /// Returns information if a pulse has been detected
    /// \return true if a pulse has been detected, false otherwise
    int isActive() const;

    /// Return the time when the last pulse was detected
    /// \return time when the last pulse was detected
    time_t getLastPulseTime() const;

//====================web server========================
#ifdef WEB_SERVER
private:
    /// A map of different values (keys) and their associated functions
    /// which returns the appropriate values.
    std::map<int, String (*)(const PulseCounter& pulseCounter)> htmlData;

public:
    /// Since this class is registered as a source of data for the
    /// HTML content, it needs return the appropriate value according 
    /// to the id given as a parameter.
    /// 
    /// \param id - the id of the piece of data
    /// \return If the class holds this piece of information, it will
    ///         return it. Otherwise, "UNDEFINED" will be returned.
    const String getHTMLData(const int id) const;

private:
    /// Associated function for returing the last pulse time
    /// \param pulseCounter - the instance of the PulseCounter class
    /// \return the last pulse time in a string format
    friend String HTML_lastPulseTime(const PulseCounter& pulseCounter);

    /// Associated function for returing the current time
    /// \param pulseCounter - the instance of the PulseCounter class
    /// \return the current time in a string format
    friend String HTML_currentTime(const PulseCounter& pulseCounter);

    /// Associated function for returing the current date
    /// \param pulseCounter - the instance of the PulseCounter class
    /// \return the current date in a string format
    friend String HTML_currentDate(const PulseCounter& pulseCounter);
#endif
};

#endif