#ifndef TOTAL_LEAK_DETECTION_H
#define TOTAL_LEAK_DETECTION_H

#include "Setup.h"
#include "HighLeakDetection.h"

//==================WebServer class=====================

/// \author silhavyj A17B0362P
///
/// This class implements the total-water leak detection
/// algorithm. It takes adventage of the already implemented
/// in the high-water leak detection algorithm and overwrites
/// some required parts of it.
class TotalLeakDetection : public HighLeakDetection {

//====================private data======================
private:
    time_t startDayPulseTime; ///< the start of the monoring period

//==================protected methods===================
protected:
    /// Returns percentage information about how close a total-water leak is from being detected
    ///
    /// The particular value is worked out based on the current settings.
    /// This piece of information is then displayed to the user.
    ///
    /// \return information about how close a total-water leak is from being detected 
    float getPercentLeakDetectionReset() const;

    /// Tests if a total-water leak has occurred, and
    /// if so, it will set the appropriate variables and flags.
    void testResetLeak() override;

    /// Tests if the total-water the algorithm should be reset
    /// if the current monitoring period is over
    void testActiveLeak() override;

//==================public methods======================
public:
    /// Constructor of the class
    ///
    /// \param pulseCounter instance of PulseCounter (input)
    /// \param type of the algorithm (total-water)
    /// \param normalConfig default normal settings
    /// \param alarmConfig default alarm settings
    TotalLeakDetection(PulseCounter *pulseCounter,
                       Type type,
                       LeakDetectionConfig_t normalConfig,
                       LeakDetectionConfig_t alarmConfig);

    /// Descrutor of the class
    ~TotalLeakDetection(){};

    /// Resets the algorithm
    void reset() override;

//====================debugging=========================
#ifdef DEBUG
    /// Returns the log id of the class
    /// \return the log id
    const String getLogID() const override;
#endif

//==================LCD display=========================
#ifdef LCD_DISPLAY
    /// Returns the content of the row given as a parameter.
    ///
    /// This class represents one page on the LCD display.
    /// Each of the 4 rows has different content identified
    /// by the number of the row.
    ///
    /// \param row number of the row
    /// \return content to be displayed on that row of the LCD
    const String getRow(int row) const override;
#endif

//====================web server========================
#ifdef WEB_SERVER
protected:
    /// A map of different values (keys) and their associated functions
    /// which returns the appropriate values.
    std::map<int, String (*)(const TotalLeakDetection& highLeakDetection)> htmlData;

public:
    /// Since this class is registered as a source of data for the
    /// HTML content, it needs return the appropriate value according 
    /// to the id given as a parameter.
    /// 
    /// \param id - the id of the piece of data
    /// \return If the class holds this piece of information, it will
    ///         return it. Otherwise, "UNDEFINED" will be returned.
    const String getHTMLData(const int id) const override;

private:
    /// Associated function for returing information about the state of the bypass
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return "OFF"/"ON" according to the current state of the bypass
    friend String HTML_bypass(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing information about a total-water leak being detected (1)
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return "ON" if a total-water leak has occurred, "OFF" otherwise
    friend String HTML_detected(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing information about a total-water leak being detected (2)
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return "Detected" if a total-water leak has occurred, "Undetected" otherwise    
    friend String HTML_detectedText(const TotalLeakDetection& totalLeakDetection);
    
    /// Associated function for returing the detection time of a total-water leak
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return the time of detection in a string format
    friend String HTML_detectedTime(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the [%] information about how close a leak is from being detected
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return [%] information about how close a leak is from being detected
    friend String HTML_activePercentage(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing a color according to [%] information about how close a leak is from being detected
    ///
    /// For instance,  0 -  59% -> gray
    ///               60 -  79% -> yellow
    ///               80 -  89% -> orange    
    ///               90 - 100% ->red
    ///
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return color according to the state of the algorithm - detected [%]
    friend String HTML_activePercentageColor(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the [%] information about how close the algorithm is from being reset
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return [%] information about how close the algorithm is from being reset
    friend String HTML_resetPercentage(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the limit (detection) in terms of the normal configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return the limit (detection) in the normal configuration
    friend String HTML_detectionLimit(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the reset time (days) in the normal configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return reset time (days) in the normal configuration
    friend String HTML_resetLimitDays(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the reset time (hours) in the normal configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return reset time (hours) in the normal configuration
    friend String HTML_resetLimitHours(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the reset time (mins) in the normal configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return reset time (mins) in the normal configuration
    friend String HTML_resetLimitMins(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the reset time (secs) in the normal configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return reset time (secs) in the normal configuration
    friend String HTML_resetLimitSecs(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the limit (detection) in terms of the alarm configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return the limit (detection) in the alarm configuration
    friend String HTML_detectionLimitAlarm(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the reset time (days) in the alarm configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return reset time (days) in the alarm configuration
    friend String HTML_resetLimitDaysAlarm(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the reset time (hours) in the alarm configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return reset time (hours) in the alarm configuration
    friend String HTML_resetLimitHoursAlarm(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the reset time (mins) in the alarm configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return reset time (mins) in the alarm configuration
    friend String HTML_resetLimitMinsAlarm(const TotalLeakDetection& totalLeakDetection);

    /// Associated function for returing the reset time (secs) in the alarm configuration
    /// \param totalLeakDetection - the instance of the TotalLeakDetection class
    /// \return reset time (secs) in the alarm configuration
    friend String HTML_resetLimitSecsAlarm(const TotalLeakDetection& totalLeakDetection);
#endif

//===============email notifications====================
#ifdef EMAIL_NOTIFICATION
private:
    /// Sends an e-mail when a total-water leak has been detected
    ///
    /// The message contains information such as the time of
    /// detection, state of the bypass button, and the limit
    /// that has been reached resulting in a total-water leak.
    void sendEmailLeakDetected() override;
#endif
};

#endif