#ifndef WEB_SERVER_H
#define WEB_SERVER_H

#include "Arduino.h"
#include "Ethernet.h"
#include "SPI.h"
#include "SD.h"
#include "vector"
#include "Setup.h"
#include "HTMLDataSource.h"
#include "IControllable.h"
#include "LeaksController.h"
#include "EmailSender.h"

#define DATA_PREF "%*" ///< start of a sequence to be replaces with an actual value
#define DATA_POST "*%" ///< end of a sequence to be replaces with an actual value

#define INDEX_FILE_NAME "INDEX~1.HTM"           ///< file on the SD card containing the HTML website
#define OUT_OF_RANGE_ERR 4294967295UL           ///< maximum integer values allowed to be passed to Arduino
#define MAX_HTTP_RQ_LENGTH 160                  ///< max size of an HTTP request (160B)
#define HTTP_RQ_GET_START "GET "                ///< the very start of t
#define HTTP_RQ_SETTINGS "/?settings="          ///< begining of an HTTP request for changing settings
#define HTTP_RQ_NOTIFICATION "/?notification="  ///< begining of an HTTP request for changing e-mail notifications
#define HTTP_RQ_SEPARATOR ';'                   ///< separator between individual parameters in an HTTP request

//==================WebServer class=====================

/// \author silhavyj A17B0362P
///
/// This class implements the functionality of the
/// webserver. This includes loading the content of
/// the website off the SD card, accepting and
/// validating HTTP requests, etc.
class WebServer : public IControllable {

//====================private data======================
private:
    static WebServer* instance;                           ///< the instance of the class 
    std::vector<HTMLDataSource*> htmlDataSources;         ///< collection of classing providing data for the HTML content      
    byte mac[6] = { 0x00, 0xAA, 0xBB, 0xCC, 0xDE, 0x02 }; ///< MAC address of the Ethernet shield
    EthernetServer server = 80;                           ///< instance of EthernetServer listening on port 80   
    File webFile;                                         ///< file   
    String buff;                                          ///< buffer for parsing an HTTP request
    LeaksController *leaksController;                     ///< instance of LeaksController so it can be updated with each line read off the file 
    bool online;                                          ///< flag if the server is up and running

//==================private methods=====================
private:
    /// Constructor of the class
    WebServer();

    /// Copy constructor of the class
    WebServer(WebServer const&){};

    /// Assignment operator of the class
    WebServer& operator=(WebServer const&) { return *this; }

    /// Inserts a piece of data into a line of the HTML content
    ///
    /// If the line contains a sequence begining with "%*" and
    /// ending with "*%". For example, "%*150*%". It will replace
    /// it with the appropriate data according to the id, 150. 
    ///
    /// \param html the current line being read off the SD card
    /// \return the same line with the piece of data interted in it
    const String replaceHTMLWithData(String html) const;

    /// Sets up the ethernet shield
    /// \return 1 if the shield has been initialized successfully, 0 otherwise
    int setEthernet();

    /// Saves settings on the SD card
    ///
    /// This method is called when new settings are applied
    /// or when the user changes the receiver's e-mail address.
    void saveSettings();

// these methods must be public when unit testing
#ifdef UNIT_TEST 
public:
#else
private:
#endif
    /// Processes an HTTP request
    ///
    /// It will determinate what kind of HTTP request has been sent
    /// and validates it afterwards. It the HTTP request does not have
    /// the right syntax, it will ber discarded.
    ///
    /// \param httpRequest the HTTP request
    /// \return 0 if the HTTP request doesn't have the right syntax, 1 otherwise
    int processHTTPRequest(String &httpRequest);

    /// Processes an HTTP request for changing settings of the water leak detection algorithms
    ///
    /// \param startPos start position (index) of the data within the HTTP request
    /// \param httpRequest the HTTP request itself
    /// \param numberOfValues number of expected parameters in the HTTP request
    /// \return 0 if the HTTP request doesn't have the right syntax, 1 otherwise
    int processHTTPRequestSettings(int startPos, String &httpRequest, int numberOfValues);

#ifdef EMAIL_NOTIFICATION
    /// Processes an HTTP request for changing settings of the e-mail notifications
    ///
    /// \param startPos start position (index) of the data within the HTTP request
    /// \param httpRequest the HTTP request itself
    /// \param numberOfValues number of expected parameters (values) in the HTTP request
    /// \return 0 if the HTTP request doesn't have the right syntax, 1 otherwise
    int processHTTPRequestNotification(int startPos, String &httpRequest, int numberOfValues);
#endif

//==================public methods======================
public:
    /// Returns the instance of the class
    ///
    /// If the instance has not been created, it will 
    /// create it and then return it.
    ///
    /// \return the instance of the class
    static WebServer* getInstance();

    /// Sets the pointer to an instance of LeaksController
    ///
    /// The purpose of this to keep updating the "low-level"
    /// part of the system as frequently as possible regardless
    /// of the webpage being loaded off the SD card. 
    ///
    /// \param leaksController instance of LeaksController
    void setup(LeaksController *leaksController);

    /// Updates the class
    void update() override;

    /// Adds another source of data for the HTML content
    /// \param source another class to be added to the collection
    void addHTMLSource(HTMLDataSource* source);
};

#endif