#include "ALeakDetectable.h"

ALeakDetectable::ALeakDetectable(PulseCounter *pulseCounter,
                                 Type type,
                                 LeakDetectionConfig_t normalConfig,
                                 LeakDetectionConfig_t alarmConfig) {

    // store all the params given through the constructor
    this->normalConfig    = normalConfig;
    this->alarmConfig     = alarmConfig;
    this->newNormalConfig = normalConfig;
    this->newAlarmConfig  = alarmConfig;
    this->type            = type;
    this->pulseCounter    = pulseCounter;
    
    // the normal settings are used by default
    config = &normalConfig;

    // initialize default values
    requestUpdateNormalConfig = 0;
    requestUpdateAlarmConfig  = 0;
    stateOfHomeAlarm          = 0;
    oldBypassValue            = 0;
}

LeakDetectionConfig_t ALeakDetectable::getAlarmConfig() {
    return alarmConfig;
}

LeakDetectionConfig_t ALeakDetectable::getNormalConfig() {
    return normalConfig;
}

void ALeakDetectable::applyNewConfig() {
    // update the normal settings
    if (requestUpdateNormalConfig) {
        #ifdef EMAIL_NOTIFICATION
            sendEmailAboutSettings(EmailSender::APPLIED_NEW_SETTING, normalConfig, newNormalConfig, false);
        #endif
        normalConfig = newNormalConfig;
        requestUpdateNormalConfig = 0;
    }
    // update the alarm settings
    if (requestUpdateAlarmConfig) {
        #ifdef EMAIL_NOTIFICATION
            sendEmailAboutSettings(EmailSender::APPLIED_NEW_SETTING, alarmConfig, newAlarmConfig, true);
        #endif
        alarmConfig = newAlarmConfig;
        requestUpdateAlarmConfig = 0;
    }
    // set the current settings according to
    // the state of the home alarm
    if (stateOfHomeAlarm)
        config = &alarmConfig;
    else config = &normalConfig;
}

void ALeakDetectable::updateNormalConfig(LeakDetectionConfig_t newNormalConfig) {
    if (this->normalConfig != newNormalConfig) { // test if the settings are not the same
        this->newNormalConfig = newNormalConfig; // store the new settings
        requestUpdateNormalConfig = 1;           // set the flag so it can be applied when the algorithm resets

        // if e-mail notifications are enabled send 
        // an e-mail to the user
        #ifdef EMAIL_NOTIFICATION
            sendEmailAboutSettings(EmailSender::CHANGED_SETTINGS, normalConfig, newNormalConfig, false);
        #endif
    }
}

void ALeakDetectable::updateAlarmConfig(LeakDetectionConfig_t newAlarmConfig) {
    if (this->alarmConfig != newAlarmConfig) { // test if the settings are not the same
        this->newAlarmConfig = newAlarmConfig; // store the new settings
        requestUpdateAlarmConfig = 1;          // set the flag so it can be applied when the algorithm resets

        // if e-mail notifications are enabled send 
        // an e-mail to the user
        #ifdef EMAIL_NOTIFICATION
            sendEmailAboutSettings(EmailSender::CHANGED_SETTINGS, alarmConfig, newAlarmConfig, true);
        #endif
    }
}

String ALeakDetectable::getType() const {
    switch (type) {
        case High:  return "High";
        case Low:   return "Low";
        case Total: return "Total";
    }
    return "unknown";
}

void ALeakDetectable::changeStateOfHomeAlarm(bool state) {
    stateOfHomeAlarm = state;
}

#ifdef EMAIL_NOTIFICATION
    void ALeakDetectable::sendEmailAboutSettings(EmailSender::Type emailType, LeakDetectionConfig_t &oldSettings, LeakDetectionConfig_t &newSettings, bool alarm) const {
        String data = "";
        if (type == HIGH || type == Total) {
            data += "current settings:\r\n";
            data += "limit [l]: " + String(PULSE_TO_LITER(oldSettings.limitPulseAction)) + " | ";
            data += "reset: " + FORMAT_TIME(oldSettings.limitResetTime) + "\r\n";
            data += "new settings:\r\n";
            data += "limit [l]: " + String(PULSE_TO_LITER(newSettings.limitPulseAction)) + " | ";
            data += "reset: " + FORMAT_TIME(newSettings.limitResetTime) + "\r\n";
        }
        else if (type == Low){
            data += "current settings:\r\n";
            data += "limit: " + FORMAT_TIME(oldSettings.limitActionTime) + " | ";
            data += "reset: " + FORMAT_TIME(oldSettings.limitResetTime) + "\r\n";
            data += "new settings:\r\n";
            data += "limit: " + FORMAT_TIME(newSettings.limitActionTime) + " | ";
            data += "reset: " + FORMAT_TIME(newSettings.limitResetTime) + "\r\n";
        }
        String subject = getType();
        subject += " leak detection settings ";
        subject += emailType == EmailSender::CHANGED_SETTINGS ? "changed " : "put in place ";
        subject += alarm ? "(ALARM)" : "(NORMAL)";

        EmailSender::getInstance()->sendEmail(emailType, subject, data);
    }
#endif