#include "DailyOverview.h"

DailyOverview::DailyOverview(HighLeakDetection *highLeakDetection,
                             LowLeakDetection *lowLeakDetection,
                             TotalLeakDetection *totalLeakDetection,
                             Consumption *dailyConsumption,
                             Consumption *monthlyConsumption) {

    // store all the params given through the constructor
    this->highLeakDetection  = highLeakDetection;   // high algorithm
    this->lowLeakDetection   = lowLeakDetection;    // low algorithm
    this->totalLeakDetection = totalLeakDetection;  // total algorithm

    this->dailyConsumption   = dailyConsumption;
    this->monthlyConsumption = monthlyConsumption;

    // store the initial date time
    initialDateTime = DateTime::getInstance()->getDateTime();

    // set the hour and minute (e.g 10:30) when the daily
    // overview should be sent to the user
    initialDateTime.hour   = DAILY_OVERVIEW_INIT_HOUR;
    initialDateTime.minute = DAILY_OVERVIEW_INIT_MIN;

    // reset the monitoring period
    reset();
}

void DailyOverview::reset() {
    maxHighLeakPercentage  = 0;
    maxLowLeakPercentage   = 0;
    maxTotalLeakPercentage = 0;
}

void DailyOverview::update() {
    // get the current date time
    RTCDateTime currentDateTime = DateTime::getInstance()->getDateTime();

    // TODO: add the minutes to the condition as well
    // so it doesn't depend only on the hour
    if (initialDateTime.day != currentDateTime.day) {
        #ifdef EMAIL_NOTIFICATION
            String data = "";
            data += "Daily water consumption: " + String(PULSE_TO_LITER(dailyConsumption->getConsumptionCount())) + "l\r\n";
            data += "Monthly water consumption: " + String(PULSE_TO_LITER(monthlyConsumption->getConsumptionCount())) + "l\r\n";
            data += "max high leak detected: " + String(maxHighLeakPercentage) + "%\r\n";
            data += "max low leak detected: " + String(maxLowLeakPercentage) + "%\r\n";
            data += "max total leak detected: " + String(maxTotalLeakPercentage) + "%\r\n";
            EmailSender::getInstance()->sendEmail(
                    EmailSender::DAILY_OVERVIEW,
                    "Daily overview",
                    data);
        #endif

        // start a new monitoring period
        initialDateTime = currentDateTime;
        reset();
    }

    // update the maximum values (% detected) of all three
    // water leak detection algorithms 
    maxHighLeakPercentage  = fmax(maxHighLeakPercentage, highLeakDetection->getPercentLeakDetected());
    maxLowLeakPercentage   = fmax(maxLowLeakPercentage, lowLeakDetection->getPercentLeakDetected());
    maxTotalLeakPercentage = fmax(maxTotalLeakPercentage, totalLeakDetection->getPercentLeakDetected());
}

#ifdef LCD_DISPLAY
    const String DailyOverview::getRow(int row) const {
        switch(row) {
            case 0:
                // 1st row on the LCD
                return String("Overview (2)");
            case 1:
                // 2nd row on the LCD
                return String("DAY CONS.=") + String(PULSE_TO_LITER(dailyConsumption->getConsumptionCount())) + "l"; 
                break;
            case 2:
                // 3rd row on the LCD
                return String("MON CONS.=") + String(PULSE_TO_LITER(monthlyConsumption->getConsumptionCount())) + "l"; 
                break;
            case 3:
                // 4th row on the LCD
                return "";
                break;
        }
        return "UNTITLED"; // for any other row return "UNTITLED"
    }
#endif