#include "EmailSender.h"

EmailSender* EmailSender::instance = NULL;

#ifdef WEB_SERVER
    String HTML_Notification_BOOTING(const EmailSender& emailSender);
    String HTML_Notification_DAILY_OVERVIEW(const EmailSender& emailSender);
    String HTML_Notification_LEAK_DETECTED(const EmailSender& emailSender);
    String HTML_Notification_VALVE_STATE(const EmailSender& emailSender);
    String HTML_Notification_RESET(const EmailSender& emailSender);
    String HTML_Notification_ALARM(const EmailSender& emailSender);
    String HTML_Notification_CHANGED_SETTINGS(const EmailSender& emailSender);
    String HTML_Notification_APPLIED_NEW_SETTING(const EmailSender& emailSender);
    String HTML_ReceiverEmailAddress(const EmailSender& emailSender);
    String HTML_Notification_BYPASS(const EmailSender& emailSender);
#endif

int EmailSender::setEthernet() {
    pinMode(SS_SD_CARD, OUTPUT);
    pinMode(SS_ETHERNET, OUTPUT);
    digitalWrite(SS_SD_CARD, HIGH);  // SD Card not active
    digitalWrite(SS_ETHERNET, HIGH); // Ethernet not active

    Ethernet.begin(mac);  // set the mac addr of the ethernet shield

    // HW not found error
    if (Ethernet.hardwareStatus() == EthernetNoHardware) {
        #ifdef DEBUG
            Serial.println("Ethernet shield was not found");
        #endif
        return 0;
    }
    // cable is not connected error
    if (Ethernet.linkStatus() == LinkOFF) {
        #ifdef DEBUG
            Serial.println("Ethernet cable is not connected.");
        #endif
        return 0;
    }
    return 1;
}

void EmailSender::enableNotification(Type type, bool state) {
    enabled[type] = state;
}

EmailSender::EmailSender() {
    if (Ethernet.linkStatus() != LinkON) {
        if (!setEthernet())
            return;
    }
    // by default disable all e-mail notifications
    enabled[BOOTING]             = false;
    enabled[DAILY_OVERVIEW]      = false;
    enabled[LEAK_DETECTED]       = false;
    enabled[VALVE_STATE]         = false;
    enabled[RESET]               = false;
    enabled[ALARM]               = false;
    enabled[CHANGED_SETTINGS]    = false;
    enabled[APPLIED_NEW_SETTING] = false;
    enabled[BYPASS]              = false;

    #ifdef WEB_SERVER
        // functions providing data to the webserver
        htmlData[160] = &HTML_Notification_BOOTING;
        htmlData[161] = &HTML_Notification_DAILY_OVERVIEW;
        htmlData[162] = &HTML_Notification_LEAK_DETECTED;
        htmlData[163] = &HTML_Notification_VALVE_STATE;
        htmlData[164] = &HTML_Notification_RESET;
        htmlData[165] = &HTML_Notification_ALARM;
        htmlData[166] = &HTML_Notification_CHANGED_SETTINGS;
        htmlData[167] = &HTML_Notification_APPLIED_NEW_SETTING;
        htmlData[170] = &HTML_ReceiverEmailAddress;
        htmlData[205] = &HTML_Notification_BYPASS;
    #endif
}

void EmailSender::sendEmail(Type type,String subject, String data) {
    if (enabled[type])
        sendEmail(subject, data);
}

void EmailSender::setReceiverEmailAddress(String email) {
    this->receiverEmailAddress = email;
}

EmailSender* EmailSender::getInstance() {
    if (instance == NULL)
        instance = new EmailSender;
    return instance;
}

byte EmailSender::sendEmail(String subject, String data) {
    // connecting to the server
    int success = client.connect(SMTPServerAddress, SMTPServerPort);
    #ifdef DEBUG
        if (success == 0)
            Serial.println("connected to the SMTPServer");
        else Serial.println("connection to the SMTPServer failed");
    #endif
    if (success != 1)
        return 0;

    if (!eRcv())
        return 0;

    // sending the "HELLO" message
    #ifdef DEBUG
        Serial.println("Sending hello");
    #endif
    client.println("EHLO " + String(arduinoIP));
    if (!eRcv())
        return 0;

    // sending a message to authenticate
    #ifdef DEBUG
        Serial.println("Sending auth login");
    #endif
    client.println("auth login");
    if (!eRcv())
        return 0;

    // sending the login
    #ifdef DEBUG
    Serial.println(("Sending User"));
    #endif
    client.println(auth64Login);
    if (!eRcv())
        return 0;

    // sending the password
    #ifdef DEBUG
        Serial.println(F("Sending Password"));
    #endif
    client.println(auth64Pass);
    if (!eRcv())
        return 0;

    // sending the sender's email address
    #ifdef DEBUG
        Serial.println("Sending From");
    #endif
    client.println("MAIL From: <" + String(emailAddress) + ">");
    if (!eRcv())
        return 0;

    // sending the receiver's email address
    #ifdef DEBUG
        Serial.println("Sending To");
    #endif
    client.println("RCPT To: <" + String(receiverEmailAddress) + ">");
    if (!eRcv())
        return 0;

    // starting sending data
    #ifdef DEBUG
        Serial.println("Sending DATA");
    #endif
    client.println("DATA");
    if (!eRcv())
        return 0;

    // sending the e-mail
    #ifdef DEBUG
        Serial.println("Sending email");
    #endif
    client.println("To: User <" + String(receiverEmailAddress) + ">");
    client.println("From: Home Water Leaks Detection <" + String(emailAddress) + ">");

    client.println("Subject: " + subject); // subject
    client.println(data);                  // body 
    client.println(".");                   // end of the e-mail 
    if (!eRcv())
        return 0;

    // log out
    #ifdef DEBUG
        Serial.println("Sending QUIT");
    #endif
    client.println("QUIT");
    if (!eRcv())
        return 0;

    // close the connection
    client.stop();
    #ifdef DEBUG
        Serial.println("disconnected");
    #endif
    return 1;
}

byte EmailSender::eRcv() {
    byte respCode;
    byte thisByte;
    int loopCount = 0;

    while (!client.available()) {
        delay(1);
        loopCount++;

        if (loopCount > 10000) {
            client.stop();
            #ifdef DEBUG
                Serial.println("\r\nTimeout");
            #endif
            return 0;
        }
    }
    respCode = client.peek();
    while (client.available()) {
        thisByte = client.read();
        #ifdef DEBUG
            Serial.write(thisByte);
        #endif
    }
    if (respCode >= '4') {
        efail();
        return 0;
    }
    return 1;
}

void EmailSender::efail() {
    byte thisByte = 0;
    int loopCount = 0;
    client.println(F("QUIT"));
    while (!client.available()) {
        delay(1);
        loopCount++;
        if (loopCount > 10000) {
            client.stop();
            #ifdef DEBUG
                Serial.println("\r\nTimeout");
            #endif
            return;
        }
    }
    while (client.available()) {
        thisByte = client.read();
        Serial.write(thisByte);
    }
    client.stop();
    #ifdef DEBUG
        Serial.println("disconnected");
    #endif
}

String EmailSender::getFormatOfSettingsToSave() {
    return String(enabled[BOOTING])             + ";" +
           String(enabled[DAILY_OVERVIEW])      + ";" +
           String(enabled[LEAK_DETECTED])       + ";" +
           String(enabled[VALVE_STATE])         + ";" +
           String(enabled[RESET])               + ";" +
           String(enabled[ALARM])               + ";" +
           String(enabled[CHANGED_SETTINGS])    + ";" +
           String(enabled[APPLIED_NEW_SETTING]) + ";" +
           String(enabled[BYPASS])              + ";" +
           receiverEmailAddress                 + ";"; // user's e-mail address
}

#ifdef WEB_SERVER
    const String EmailSender::getHTMLData(const int id) const {
        auto fce = htmlData.find(id);
        if (fce == htmlData.end())
            return HTMLDataSource::UNDEFINED_DATA;
        return fce->second(*this);
    }

    String HTML_Notification_BOOTING(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::BOOTING);
        return it->second == true ? "checked" : "";
    }

    String HTML_Notification_DAILY_OVERVIEW(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::DAILY_OVERVIEW);
        return it->second == true ? "checked" : "";
    }

    String HTML_Notification_LEAK_DETECTED(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::LEAK_DETECTED);
        return it->second == true ? "checked" : "";
    }

    String HTML_Notification_VALVE_STATE(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::VALVE_STATE);
        return it->second == true ? "checked" : "";
    }

    String HTML_Notification_RESET(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::RESET);
        return it->second == true ? "checked" : "";
    }

    String HTML_Notification_ALARM(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::ALARM);
        return it->second == true ? "checked" : "";
    }

    String HTML_Notification_CHANGED_SETTINGS(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::CHANGED_SETTINGS);
        return it->second == true ? "checked" : "";
    }

    String HTML_Notification_APPLIED_NEW_SETTING(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::APPLIED_NEW_SETTING);
        return it->second == true ? "checked" : "";
    }

    String HTML_Notification_BYPASS(const EmailSender& emailSender) {
        auto it = emailSender.enabled.find(EmailSender::BYPASS);
        return it->second == true ? "checked" : "";
    }

    String HTML_ReceiverEmailAddress(const EmailSender& emailSender) {
        return emailSender.receiverEmailAddress;
    }
#endif