#include "HighLeakDetection.h"

#ifdef WEB_SERVER
    String HTML_bypass(const HighLeakDetection& highLeakDetection);
    String HTML_detected(const HighLeakDetection& highLeakDetection);
    String HTML_activePercentage(const HighLeakDetection& highLeakDetection);
    String HTML_resetPercentage(const HighLeakDetection& highLeakDetection);
    String HTML_detectedText(const HighLeakDetection& highLeakDetection);
    String HTML_activePercentageColor(const HighLeakDetection& highLeakDetection);
    String HTML_detectedTime(const HighLeakDetection& highLeakDetection);
    String HTML_detectionLimit(const HighLeakDetection& highLeakDetection);
    String HTML_resetLimitDays(const HighLeakDetection& highLeakDetection);
    String HTML_resetLimitHours(const HighLeakDetection& highLeakDetection);
    String HTML_resetLimitMins(const HighLeakDetection& highLeakDetection);
    String HTML_resetLimitSecs(const HighLeakDetection& highLeakDetection);
    String HTML_detectionLimitAlarm(const HighLeakDetection& highLeakDetection);
    String HTML_resetLimitDaysAlarm(const HighLeakDetection& highLeakDetection);
    String HTML_resetLimitHoursAlarm(const HighLeakDetection& highLeakDetection);
    String HTML_resetLimitMinsAlarm(const HighLeakDetection& highLeakDetection);
    String HTML_resetLimitSecsAlarm(const HighLeakDetection& highLeakDetection);
#endif

HighLeakDetection::HighLeakDetection(PulseCounter *pulseCounter,
                                     Type type,
                                     LeakDetectionConfig_t normalConfig,
                                     LeakDetectionConfig_t alarmConfig)
        : ALeakDetectable(pulseCounter,type, normalConfig, alarmConfig) {
    #ifdef WEB_SERVER
        // functions providing data to the webserver
        htmlData[0]  = &HTML_bypass;
        htmlData[1]  = &HTML_detected;
        htmlData[2]  = &HTML_activePercentage;
        htmlData[3]  = &HTML_resetPercentage;
        htmlData[20] = &HTML_detectedText;
        htmlData[21] = &HTML_activePercentageColor;
        htmlData[26] = &HTML_detectedTime;

        htmlData[40] = &HTML_detectionLimit;
        htmlData[41] = &HTML_resetLimitDays;
        htmlData[42] = &HTML_resetLimitHours;
        htmlData[43] = &HTML_resetLimitMins;
        htmlData[44] = &HTML_resetLimitSecs;

        htmlData[50] = &HTML_detectionLimitAlarm;
        htmlData[51] = &HTML_resetLimitDaysAlarm;
        htmlData[52] = &HTML_resetLimitHoursAlarm;
        htmlData[53] = &HTML_resetLimitMinsAlarm;
        htmlData[54] = &HTML_resetLimitSecsAlarm;
    #endif
    reset(); // reset the algorithm
}

void HighLeakDetection::reset() {
    active = LOW;
    bypass = LOW;
    detected          = 0;
    resetWhenDetected = 0;
    pulseCount        = 0;
    timeOfDetection = "";
    ALeakDetectable::applyNewConfig(); // check if new settings should be applied
}

String HighLeakDetection::getFormatOfSettingsToSave() {
    return String(newNormalConfig.limitResetTime)   + ";" +
           String(newNormalConfig.limitPulseAction) + ";" +
           String(newAlarmConfig.limitResetTime)    + ";" +
           String(newAlarmConfig.limitPulseAction)  + ";";
}

void HighLeakDetection::testResetLeak() {
    time_t currentTime   = millis(); 
    time_t lastPulseTime = pulseCounter->getLastPulseTime();
    time_t deltaTime     = DELTA_TIME(currentTime, lastPulseTime);

    if (deltaTime >= config->limitResetTime && !detected) {
        pulseCount = 0;
        ALeakDetectable::applyNewConfig();
    }
}

void HighLeakDetection::testActiveLeak() {
    detected = (pulseCount >= config->limitPulseAction);

    if (detected) {
        // store the reset % when a leak was detected
        // if it has not been done yet
        if (timeOfDetection == "") {
            timeOfDetection   = DateTime::getInstance()->getDateTimeStr(); // date time of detection
            resetWhenDetected = (int)getPercentLeakDetectionReset();       // reset % 

            // send e-mail to the user
            #ifdef EMAIL_NOTIFICATION
                sendEmailLeakDetected();
            #endif      
        }
        active = !bypass; // if the bypass is off, close the main valve
    }
    // indication LED
    digitalWrite(HIGH_ILOCK_LED_PIN, detected);
}

#ifdef EMAIL_NOTIFICATION
    void HighLeakDetection::sendEmailLeakDetected() {
        EmailSender::getInstance()->sendEmail(
                EmailSender::LEAK_DETECTED,
                "High leak detected!",
                "High leak was just detected\n\r"
                "time of detection: " + timeOfDetection + "; "
                "limit for leak to be detected: " +
                PULSE_TO_LITER(normalConfig.limitPulseAction) + "l | " +
                PULSE_TO_LITER(alarmConfig.limitPulseAction) + "l (alarm)");
    }
#endif

void HighLeakDetection::update() {
    // read the bypass
    bypass = digitalRead(config->bypassPin);

    // if the state of the bypass has changed
    // send an e-mail to the user
    #ifdef EMAIL_NOTIFICATION
        if (oldBypassValue != bypass) {
            oldBypassValue = bypass;
            String state = bypass ? "on" : "off";
            EmailSender::getInstance()->sendEmail(
                    EmailSender::BYPASS,
                    getType() + " leak detection BYPASS",
                    "bypass is  now " + state
            );
        }
    #endif

    // if a pulse has been detected and a leak 
    // has not been detected, increment the counter
    if (pulseCounter->isActive() && detected == LOW)
        pulseCount++;

    testActiveLeak(); // test if a leak has occurred
    testResetLeak();  // test if the algorithm should be reset
}

float HighLeakDetection::getPercentLeakDetected() const {
    float percentDetected = pulseCount / (float)config->limitPulseAction;
    percentDetected *= 100;
    return min(100, percentDetected);
}

float HighLeakDetection::getPercentLeakDetectionReset() const {
    if (active == HIGH)
        return 0;

    time_t currentTime = millis();
    time_t lastPulseTime = pulseCounter->getLastPulseTime();
    time_t deltaTime = DELTA_TIME(currentTime, lastPulseTime);

    float percentReset = deltaTime / (float)config->limitResetTime;
    percentReset *= 100;
    return min(100, percentReset);
}

#ifdef DEBUG
    const String HighLeakDetection::getLogID() const { return "[HIGH LEAK]"; }

    const String HighLeakDetection::getLogDescription() const {
        return "[" +
                String("BYPASS=") + String(bypass) + " | " +
                String("ACTIVE=") + String(active) + " | " +
                String("ACTIVE=") + String(getPercentLeakDetected()) + "% | " +
                String("RESET=") + String(getPercentLeakDetectionReset()) + "% | " +
                String("CONFIG=") + ((config == &alarmConfig) ? String("ALARM") : String("NORMAL")) + " | " +
                String("PULSE_COUNT=") + String(pulseCount) +
            "]";
    }
#endif

#ifdef LCD_DISPLAY
    const String HighLeakDetection::getRow(int row) const {
        switch(row) {
            case 0:
                // 1st row on the LCD
                return getType() + " water leak";
                break;
            case 1:
                // 2nd row on the LCD
                return String("BYPASS=") + String(bypass) + String(" | ") + String("ACTIVE=") + String(active);
                break;
            case 2:
                // 3rd row on the LCD
                return String("ACTIVE=") + String(getPercentLeakDetected()) + "%";
                break;
            case 3:
                // 4th row on the LCD
                return String("RESET=") + String(getPercentLeakDetectionReset()) + "%";
                break;
            break;
        }
        return "UNTITLED"; // for any other row return "UNTITLED"
    }
#endif


#ifdef WEB_SERVER
    const String HighLeakDetection::getHTMLData(const int id) const {
        auto fce = htmlData.find(id);
        if (fce == htmlData.end())
            return HTMLDataSource::UNDEFINED_DATA;
        return fce->second(*this);
    }

    String HTML_bypass(const HighLeakDetection& highLeakDetection) {
        if (highLeakDetection.bypass == LOW)
            return String("OFF");
        return String("ON");
    }

    String HTML_detectedText(const HighLeakDetection& highLeakDetection) {
        if (highLeakDetection.active)
            return String("Detected");
        return String("Undetected");
    }

    String HTML_detected(const HighLeakDetection& highLeakDetection) {
        if (highLeakDetection.active)
            return String("ON");
        return String("OFF");
    }

    String HTML_activePercentage(const HighLeakDetection& highLeakDetection) {
        return String((int)highLeakDetection.getPercentLeakDetected());
    }

    String HTML_resetPercentage(const HighLeakDetection& highLeakDetection) {
        if (highLeakDetection.detected)
            return String(highLeakDetection.resetWhenDetected);
        return String((int)highLeakDetection.getPercentLeakDetectionReset());
    }

    String HTML_activePercentageColor(const HighLeakDetection& highLeakDetection) {
        int percentage = (int)highLeakDetection.getPercentLeakDetected();
        if (percentage < 60) return "rgba(44, 62, 80, 0.301)";
        if (percentage >= 60 && percentage < 75) return "#ffdc52";
        if (percentage >= 75 && percentage < 85) return "#fca758";
        return "#fc5858";
    }

    String HTML_detectedTime(const HighLeakDetection& highLeakDetection) {
        return highLeakDetection.timeOfDetection;
    }

    String HTML_detectionLimit(const HighLeakDetection& highLeakDetection) {
        return String(PULSE_TO_LITER(highLeakDetection.normalConfig.limitPulseAction));
    }

    String HTML_resetLimitDays(const HighLeakDetection& highLeakDetection) {
        return String(DAYS(highLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_resetLimitHours(const HighLeakDetection& highLeakDetection) {
        return String(HOURS(highLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_resetLimitMins(const HighLeakDetection& highLeakDetection) {
        return String(MINS(highLeakDetection.normalConfig.limitResetTime));
    }
    String HTML_resetLimitSecs(const HighLeakDetection& highLeakDetection) {
        return String(SECS(highLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_detectionLimitAlarm(const HighLeakDetection& highLeakDetection) {
        return String(PULSE_TO_LITER(highLeakDetection.alarmConfig.limitPulseAction));
    }

    String HTML_resetLimitDaysAlarm(const HighLeakDetection& highLeakDetection) {
        return String(DAYS(highLeakDetection.alarmConfig.limitResetTime));
    }
    String HTML_resetLimitHoursAlarm(const HighLeakDetection& highLeakDetection) {
        return String(HOURS(highLeakDetection.alarmConfig.limitResetTime));
    }

    String HTML_resetLimitMinsAlarm(const HighLeakDetection& highLeakDetection) {
        return String(MINS(highLeakDetection.alarmConfig.limitResetTime));
    }

    String HTML_resetLimitSecsAlarm(const HighLeakDetection& highLeakDetection) {
        return String(SECS(highLeakDetection.alarmConfig.limitResetTime));
    }
#endif