#include "LeaksController.h"

#ifdef WEB_SERVER
    String HTML_valve(const LeaksController& leaksController);
    String HTML_alarm(const LeaksController& leaksController);
    String HTML_manualClose(const LeaksController& leaksController);
    String HTML_dailyConsumption(const LeaksController& leaksController);
    String HTML_monthlyConsumption(const LeaksController& leaksController);
#endif

LeaksController* LeaksController::instance = NULL;

LeaksController::LeaksController() {
    valveState       = LOW;
    stateOfHomeAlarm = LOW;

    #ifdef WEB_SERVER
        // functions providing data to the webserver
        htmlData[13] = &HTML_valve;
        htmlData[14] = &HTML_alarm;
        htmlData[15] = &HTML_manualClose;
        htmlData[17] = &HTML_dailyConsumption;
        htmlData[18] = &HTML_monthlyConsumption;
    #endif
}

LeaksController* LeaksController::getInstance() {
    if (instance == NULL)
        instance = new LeaksController;
    return instance;
}

void LeaksController::addLeakDetection(ALeakDetectable::Type type, ALeakDetectable* leak) {
    leaks[type] = leak;
}

int LeaksController::getNumberOfLeakDetections() const {
    return leaks.size();
}

#ifdef WEB_SERVER
    void LeaksController::setDailyConsumptionCounter(Consumption *dailyConsumptionCounter) {
        this->dailyConsumptionCounter = dailyConsumptionCounter;
    }

    void LeaksController::setMonthlyConsumptionCounter(Consumption *monthlyConsumptionCounter) {
        this->monthlyConsumptionCounter = monthlyConsumptionCounter;
    }
#endif

void LeaksController::changeSettings(ALeakDetectable::Type type, unsigned long detected, unsigned long reset, bool alarm) {
    // tests if that kind of water leak
    // detection algorithm exists
    auto it = leaks.find(type);
    if (it == leaks.end())
        return;

    // get the current settings (makes a copy
    // of the original one)
    ALeakDetectable *leak = it->second;
    LeakDetectionConfig_t config = alarm ? leak->getAlarmConfig() : leak->getNormalConfig();

    // high-water leak detection algorithm
    if (type == ALeakDetectable::High) {
        config.limitPulseAction = detected; // limit
        config.limitResetTime   = reset;    // reset
    }
    // low-water leak detection algorithm
    else if (type == ALeakDetectable::Low) {
        config.limitActionTime = detected; // limit
        config.limitResetTime  = reset;    // reset
    }
    // total-water leak detection algorithm
    else if (type == ALeakDetectable::Total) {
        config.limitPulseAction = detected; // limit
        config.limitResetTime   = reset;    // reset
    }
    // Update the settings
    if (alarm)
        leak->updateAlarmConfig(config);
    else leak->updateNormalConfig(config);
}

String LeaksController::getFormatOfSettingsToSave() {
    return leaks[ALeakDetectable::High]->getFormatOfSettingsToSave() + "\n" +
           leaks[ALeakDetectable::Low]->getFormatOfSettingsToSave()  + "\n" +
           leaks[ALeakDetectable::Total]->getFormatOfSettingsToSave();
}

void LeaksController::setPulseCounter(PulseCounter *pulseCounter) {
    this->pulseCounter = pulseCounter;
}

void LeaksController::update() {
    pulseCounter->update(); // update the pulse counter
    manualValveClose        = digitalRead(VALVE_MANUAL_CLOSE_PIN); // manual close reset button
    int stateOfHomeAlarmTmp = digitalRead(HOME_ALARM_PIN);         // read the state of the home alarm
    
    // if the state of the home alarm changed,
    // send an e-mail to the user
    #ifdef EMAIL_NOTIFICATION
        if (stateOfHomeAlarmTmp != stateOfHomeAlarm) {
            String state = stateOfHomeAlarmTmp ? "ACTIVE" : "NOT ACTIVE";
            EmailSender::getInstance()->sendEmail(
                EmailSender::ALARM,
                "Home alarm",
                "Home alarm is now " + state + "\r\n"
                "time: " + DateTime::getInstance()->getDateTimeStr()
            );
        }
    #endif
    stateOfHomeAlarm = stateOfHomeAlarmTmp;

    #ifdef WEB_SERVER
        // update the daily and month water consumptions
        dailyConsumptionCounter->update();
        monthlyConsumptionCounter->update();
    #endif

    // check if any of the algorithms has detected a water leak
    int isAnyLeakActive = 0;
    for (const auto leak : leaks) {
        // if the state of the home alarm changed
        // notify all the water leak detection algorithms
        #ifdef HOME_ALARM_SETTINGS
            leak.second->changeStateOfHomeAlarm(stateOfHomeAlarm);
        #endif

        leak.second->update();
        if (leak.second->isActive())
            isAnyLeakActive = 1;
    }

    // check out if the state of the main valve should be changed
    if ((isAnyLeakActive || manualValveClose) && valveState == LOW) {
        valveState = HIGH;
    } else if (!isAnyLeakActive && !manualValveClose && valveState == HIGH)
        valveState = LOW;

    // check if the reset button is being presset
    if (manualResetButton.isPressed()) {
        // reset all the algorithms
        for (const auto leak : leaks)
            leak.second->reset();

        // if the valve has not been closed
        // manually by the user, open it up again
        if (!manualValveClose)
            valveState = LOW;

        // send a notification e-mail to the user
        #ifdef EMAIL_NOTIFICATION
            EmailSender::getInstance()->sendEmail(
                    EmailSender::RESET,
                    "Reset of the device",
                    "All the detection algorithms have been reset");
        #endif
    }

    digitalWrite(VALVE_PIN, !valveState);    // send a signal to the main valve (0=close, 1=open)
    digitalWrite(VALVE_LED_PIN, valveState); // send a signal to the LED of the main valve
    
    #ifdef EMAIL_NOTIFICATION
        // if the state of the main valve's changed
        // send an e-mail to the user
        if (valveState != oldValveState) {
            String state, msg;
            if (valveState) {
                state = "CLOSED";
                msg   = "open to closed";
            } else {
                state = "OPEN";
                msg   = "closed to open";
            }
            EmailSender::getInstance()->sendEmail(
                    EmailSender::VALVE_STATE,
                    "Valve is now " + state,
                    "The state of the valve has been changed from " + msg);
        }
    #endif
}

#ifdef DEBUG
    const String LeaksController::getLogID() const { return "[LEAKS_CONTROLLER]"; }

    const String LeaksController::getLogDescription() const {
        return "[" +
                String("VALVE_STATE=") + String(valveState) + " | " +
                String("HOME ALARM=") + String(stateOfHomeAlarm) + " | " +
                String("MANUAL_CLOSE=") + String(manualValveClose) +
            "]";
    }
#endif

#ifdef LCD_DISPLAY
    const String LeaksController::getRow(int row) const {
        switch(row) {
            case 0:
                // 1st row on the LCD
                return String("Overview (1)");
            case 1:
                // 2nd row on the LCD
                return String("VALVE STATE=") + String(valveState); 
                break;
            case 2:
                // 3rd row on the LCD
                return String("HOME ALARM=") + String(stateOfHomeAlarm);
                break;
            case 3:
                // 4th row on the LCD
                return String("MANUAL CLOSE=") + String(manualValveClose);
                break;
        }
        return "UNTITLED"; // for any other row return "UNTITLED"
    }
#endif

#ifdef WEB_SERVER
    const String LeaksController::getHTMLData(const int id) const {
        auto fce = htmlData.find(id);
        if (fce == htmlData.end())
            return HTMLDataSource::UNDEFINED_DATA;
        return fce->second(*this);
    }

    String HTML_valve(const LeaksController& leaksController) {
        if (leaksController.valveState == HIGH)
            return String("OFF");
        return String("ON");
    }

    String HTML_alarm(const LeaksController& leaksController) {
        if (leaksController.stateOfHomeAlarm)
            return String("ON");
        return String("OFF");
    }

    String HTML_manualClose(const LeaksController& leaksController) {
        if (leaksController.manualValveClose)
            return String("ON");
        return String("OFF");
    }

    String HTML_dailyConsumption(const LeaksController& leaksController) {
        int count = leaksController.dailyConsumptionCounter->getConsumptionCount();
        return String(PULSE_TO_LITER(count));
    }

    String HTML_monthlyConsumption(const LeaksController& leaksController) {
        int count = leaksController.monthlyConsumptionCounter->getConsumptionCount();
        return String(PULSE_TO_LITER(count));
    }
#endif