#include "LowLeakDetection.h"

#ifdef WEB_SERVER
    String HTML_bypass(const LowLeakDetection& lowLeakDetection);
    String HTML_detected(const LowLeakDetection& lowLeakDetection);
    String HTML_activePercentage(const LowLeakDetection& lowLeakDetection);
    String HTML_resetPercentage(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedText(const LowLeakDetection& lowLeakDetection);
    String HTML_activePercentageColor(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedTime(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedLimitDays(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedLimitHours(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedLimitMins(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedLimitSecs(const LowLeakDetection& lowLeakDetection);
    String HTML_resetLimitDays(const LowLeakDetection& lowLeakDetection);
    String HTML_resetLimitHours(const LowLeakDetection& lowLeakDetection);
    String HTML_resetLimitMins(const LowLeakDetection& lowLeakDetection);
    String HTML_resetLimitSecs(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedLimitDaysAlarm(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedLimitHoursAlarm(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedLimitMinsAlarm(const LowLeakDetection& lowLeakDetection);
    String HTML_detectedLimitSecsAlarm(const LowLeakDetection& lowLeakDetection);
    String HTML_resetLimitDaysAlarm(const LowLeakDetection& lowLeakDetection);
    String HTML_resetLimitHoursAlarm(const LowLeakDetection& lowLeakDetection);
    String HTML_resetLimitMinsAlarm(const LowLeakDetection& lowLeakDetection);
    String HTML_resetLimitSecsAlarm(const LowLeakDetection& lowLeakDetection);
#endif

LowLeakDetection::LowLeakDetection(PulseCounter *pulseCounter,
                                   Type type,
                                   LeakDetectionConfig_t normalConfig,
                                   LeakDetectionConfig_t alarmConfig)
        : ALeakDetectable(pulseCounter, type, normalConfig, alarmConfig) {
#ifdef WEB_SERVER
    // functions providing data to the webserver
    htmlData[4]  = &HTML_bypass;
    htmlData[5]  = &HTML_detected;
    htmlData[6]  = &HTML_activePercentage;
    htmlData[7]  = &HTML_resetPercentage;
    htmlData[22] = &HTML_detectedText;
    htmlData[23] = &HTML_activePercentageColor;
    htmlData[27] = &HTML_detectedTime;

    htmlData[60] = &HTML_detectedLimitDays;
    htmlData[61] = &HTML_detectedLimitHours;
    htmlData[62] = &HTML_detectedLimitMins;
    htmlData[63] = &HTML_detectedLimitSecs;

    htmlData[70] = &HTML_resetLimitDays;
    htmlData[71] = &HTML_resetLimitHours;
    htmlData[72] = &HTML_resetLimitMins;
    htmlData[73] = &HTML_resetLimitSecs;

    htmlData[80] = &HTML_detectedLimitDaysAlarm;
    htmlData[81] = &HTML_detectedLimitHoursAlarm;
    htmlData[82] = &HTML_detectedLimitMinsAlarm;
    htmlData[83] = &HTML_detectedLimitSecsAlarm;

    htmlData[90] = &HTML_resetLimitDaysAlarm;
    htmlData[91] = &HTML_resetLimitHoursAlarm;
    htmlData[92] = &HTML_resetLimitMinsAlarm;
    htmlData[93] = &HTML_resetLimitSecsAlarm;
#endif
    reset(); // reset the algorithm
}

void LowLeakDetection::reset() {
    bypass   = LOW;
    active   = LOW;
    flipFlop = LOW;
    flipFlopSetTime   = 0;
    detected          = 0;
    resetWhenDetected = 0;
    timeOfDetection = "";
    ALeakDetectable::applyNewConfig(); // check if new settings should be applied
}

void LowLeakDetection::update() {
    // read the bypass
    bypass = digitalRead(config->bypassPin);

    // if the state of the bypass has changed
    // send an e-mail to the user
    #ifdef EMAIL_NOTIFICATION
        if (oldBypassValue != bypass) {
            oldBypassValue = bypass;
            String state = bypass ? "on" : "off";
            EmailSender::getInstance()->sendEmail(
                    EmailSender::BYPASS,
                    getType() + " leak detection BYPASS",
                    "bypass is  now " + state
            );
        }
    #endif

    // if a pulse has been detected and the flipflop
    // is down, set it up and store the time when it
    // has been set up
    if (pulseCounter->isActive() && flipFlop == LOW) {
        flipFlop = HIGH;
        flipFlopSetTime = pulseCounter->getLastPulseTime();
    }
    testActiveLeak(); // test if a leak has occurred
    testResetLeak();  // test if the algorithm should be reset
}

String LowLeakDetection::getFormatOfSettingsToSave() {
    return String(newNormalConfig.limitResetTime)  + ";" +
           String(newNormalConfig.limitActionTime) + ";" +
           String(newAlarmConfig.limitResetTime)   + ";" +
           String(newAlarmConfig.limitActionTime)  + ";";
}

void LowLeakDetection::testResetLeak() {
    time_t currentTime   = millis();
    time_t lastPulseTime = pulseCounter->getLastPulseTime();
    time_t deltaTime     = DELTA_TIME(currentTime, lastPulseTime);

    if (deltaTime >= config->limitResetTime && !detected) {
        flipFlop = LOW;
        ALeakDetectable::applyNewConfig();
    }
}

void LowLeakDetection::testActiveLeak() {
    time_t currentTime = millis();
    time_t deltaTime   = DELTA_TIME(currentTime, flipFlopSetTime);         
    detected = (flipFlop == HIGH && deltaTime >= config->limitActionTime); 

    if (detected) {
        // store the reset % when a leak was detected
        // if it has not been done yet
        if (timeOfDetection == "") {
            timeOfDetection = DateTime::getInstance()->getDateTimeStr(); // date time of detection
            resetWhenDetected = (int)getPercentLeakDetectionReset();     // reset % 

            // send e-mail to the user
            #ifdef EMAIL_NOTIFICATION
                sendEmailLeakDetected();
            #endif
        }
        active = !bypass; // if the bypass is off, close the main valve
    }
    // indication LED
    digitalWrite(LOW_ILOCK_LED_PIN, detected);
}

#ifdef EMAIL_NOTIFICATION
    void LowLeakDetection::sendEmailLeakDetected() {
        EmailSender::getInstance()->sendEmail(
                EmailSender::LEAK_DETECTED,
                "Low leak detected!",
                "Low leak was just detected\n\r"
                "time of detection: " + timeOfDetection + "; "
                "limit for leak to be detected: " +
                FORMAT_TIME(normalConfig.limitActionTime) + "l | " +
                FORMAT_TIME(alarmConfig.limitActionTime) + "l (alarm)");
    }
#endif

float LowLeakDetection::getPercentLeakDetectionReset() const {
    if (active == HIGH)
        return 0;

    time_t currentTime = millis();
    time_t lastPulseTime = pulseCounter->getLastPulseTime();
    time_t deltaTime = DELTA_TIME(currentTime, lastPulseTime);

    float percentReset = deltaTime / (float)config->limitResetTime;
    percentReset *= 100;
    return min(100, percentReset);
}

float LowLeakDetection::getPercentLeakDetected() const {
    if (flipFlop == LOW)
        return 0;

    time_t currentTime = millis();
    time_t deltaTime = DELTA_TIME(currentTime, flipFlopSetTime);
    float percentDetected = deltaTime / (float)config->limitActionTime;
    percentDetected *= 100;
    return min(100, percentDetected);
}

#ifdef DEBUG
    const String LowLeakDetection::getLogID() const { return "[LOW LEAK]"; }

    const String LowLeakDetection::getLogDescription() const {
        return "[" +
            String("BYPASS=") + String(bypass) + " | " +
            String("ACTIVE=") + String(active) + " | " +
            String("ACTIVE=") + String(getPercentLeakDetected()) + "% | " +
            String("RESET=") + String(getPercentLeakDetectionReset()) + "% | " +
            String("FLIP_FLOP=") + String(flipFlop) + " | " +
            String("FLIP_FLOP_SET_TIME=") + FORMAT_TIME(flipFlopSetTime) + " | " +
            String("CONFIG=") + ((config == &alarmConfig) ? String("ALARM") : String("NORMAL")) +
                "]";
    }
#endif

#ifdef LCD_DISPLAY
    const String LowLeakDetection::getRow(int row) const {
        switch(row) {
            case 0:
                // 1st row on the LCD
                return getType() + " water leak";
                break;
            case 1:
                // 2nd row on the LCD
                return String("BYPASS=") + String(bypass) + String(" | ") + String("ACTIVE=") + String(active);
                break;
            case 2:
                // 3rd row on the LCD
                return String("ACTIVE=") + String(getPercentLeakDetected()) + "%";
                break;
            case 3:
                // 4th row on the LCD
                return String("RESET=") + String(getPercentLeakDetectionReset()) + "%";
                break;
            break;
        }
        return "UNTITLED"; // for any other row return "UNTITLED"
    }
#endif

#ifdef WEB_SERVER
    const String LowLeakDetection::getHTMLData(const int id) const {
        auto fce = htmlData.find(id);
        if (fce == htmlData.end())
            return HTMLDataSource::UNDEFINED_DATA;
        return fce->second(*this);
    }

    String HTML_detectedText(const LowLeakDetection& lowLeakDetection) {
        if (lowLeakDetection.active == LOW)
            return String("Undetected");
        return String("Detected");
    }

    String HTML_bypass(const LowLeakDetection& lowLeakDetection) {
        if (lowLeakDetection.bypass == LOW)
            return String("OFF");
        return String("ON");
    }

    String HTML_detected(const LowLeakDetection& lowLeakDetection) {
        if (lowLeakDetection.active)
            return String("ON");
        return String("OFF");
    }

    String HTML_activePercentage(const LowLeakDetection& lowLeakDetection) {
        return String((int)lowLeakDetection.getPercentLeakDetected());
    }

    String HTML_resetPercentage(const LowLeakDetection& lowLeakDetection) {
        if (lowLeakDetection.detected)
            return String(lowLeakDetection.resetWhenDetected);
        return String((int)lowLeakDetection.getPercentLeakDetectionReset());
    }

    String HTML_activePercentageColor(const LowLeakDetection& lowLeakDetection) {
        int percentage = (int)lowLeakDetection.getPercentLeakDetected();
        if (percentage < 60) return "rgba(44, 62, 80, 0.301)";
        if (percentage >= 60 && percentage < 75) return "#ffdc52";
        if (percentage >= 75 && percentage < 85) return "#fca758";
        return "#fc5858";
    }

    String HTML_detectedTime(const LowLeakDetection& lowLeakDetection) {
        return lowLeakDetection.timeOfDetection;
    }

    String HTML_detectedLimitDays(const LowLeakDetection& lowLeakDetection) {
        return String(DAYS(lowLeakDetection.normalConfig.limitActionTime));
    }

    String HTML_detectedLimitHours(const LowLeakDetection& lowLeakDetection) {
        return String(HOURS(lowLeakDetection.normalConfig.limitActionTime));
    }

    String HTML_detectedLimitMins(const LowLeakDetection& lowLeakDetection) {
        return String(MINS(lowLeakDetection.normalConfig.limitActionTime));
    }

    String HTML_detectedLimitSecs(const LowLeakDetection& lowLeakDetection) {
        return String(SECS(lowLeakDetection.normalConfig.limitActionTime));
    }

    String HTML_resetLimitDays(const LowLeakDetection& lowLeakDetection) {
        return String(DAYS(lowLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_resetLimitHours(const LowLeakDetection& lowLeakDetection) {
        return String(HOURS(lowLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_resetLimitMins(const LowLeakDetection& lowLeakDetection) {
        return String(MINS(lowLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_resetLimitSecs(const LowLeakDetection& lowLeakDetection) {
        return String(SECS(lowLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_detectedLimitDaysAlarm(const LowLeakDetection& lowLeakDetection) {
        return String(DAYS(lowLeakDetection.alarmConfig.limitActionTime));
    }

    String HTML_detectedLimitHoursAlarm(const LowLeakDetection& lowLeakDetection) {
        return String(HOURS(lowLeakDetection.alarmConfig.limitActionTime));
    }
    String HTML_detectedLimitMinsAlarm(const LowLeakDetection& lowLeakDetection) {
        return String(MINS(lowLeakDetection.alarmConfig.limitActionTime));
    }

    String HTML_detectedLimitSecsAlarm(const LowLeakDetection& lowLeakDetection) {
        return String(SECS(lowLeakDetection.alarmConfig.limitActionTime));
    }

    String HTML_resetLimitDaysAlarm(const LowLeakDetection& lowLeakDetection) {
        return String(DAYS(lowLeakDetection.alarmConfig.limitResetTime));
    }

    String HTML_resetLimitHoursAlarm(const LowLeakDetection& lowLeakDetection) {
        return String(HOURS(lowLeakDetection.alarmConfig.limitResetTime));
    }

    String HTML_resetLimitMinsAlarm(const LowLeakDetection& lowLeakDetection) {
        return String(MINS(lowLeakDetection.alarmConfig.limitResetTime));
    }

    String HTML_resetLimitSecsAlarm(const LowLeakDetection& lowLeakDetection) {
        return String(SECS(lowLeakDetection.alarmConfig.limitResetTime));
    }
#endif
