#include "TotalLeakDetection.h"

#ifdef WEB_SERVER
    String HTML_bypass(const TotalLeakDetection& totalLeakDetection);
    String HTML_detected(const TotalLeakDetection& totalLeakDetection);
    String HTML_activePercentage(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetPercentage(const TotalLeakDetection& totalLeakDetection);
    String HTML_detectedText(const TotalLeakDetection& totalLeakDetection);
    String HTML_activePercentageColor(const TotalLeakDetection& totalLeakDetection);
    String HTML_detectedTime(const TotalLeakDetection& totalLeakDetection);
    String HTML_detectionLimit(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetLimitDays(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetLimitHours(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetLimitMins(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetLimitSecs(const TotalLeakDetection& totalLeakDetection);
    String HTML_detectionLimitAlarm(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetLimitDaysAlarm(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetLimitHoursAlarm(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetLimitMinsAlarm(const TotalLeakDetection& totalLeakDetection);
    String HTML_resetLimitSecsAlarm(const TotalLeakDetection& totalLeakDetection);
#endif

TotalLeakDetection::TotalLeakDetection(PulseCounter *pulseCounter,
                                       Type type,
                                       LeakDetectionConfig_t normalConfig,
                                       LeakDetectionConfig_t alarmConfig)
        : HighLeakDetection(pulseCounter, type, normalConfig, alarmConfig) {
#ifdef WEB_SERVER
    // functions providing data to the webserver
    htmlData[8]   = &HTML_bypass;
    htmlData[9]   = &HTML_detected;
    htmlData[10]  = &HTML_activePercentage;
    htmlData[11]  = &HTML_resetPercentage;
    htmlData[24]  = &HTML_detectedText;
    htmlData[25]  = &HTML_activePercentageColor;
    htmlData[28]  = &HTML_detectedTime;

    htmlData[100] = &HTML_detectionLimit;
    htmlData[101] = &HTML_resetLimitDays;
    htmlData[102] = &HTML_resetLimitHours;
    htmlData[103] = &HTML_resetLimitMins;
    htmlData[104] = &HTML_resetLimitSecs;

    htmlData[105] = &HTML_detectionLimitAlarm;
    htmlData[106] = &HTML_resetLimitDaysAlarm;
    htmlData[107] = &HTML_resetLimitHoursAlarm;
    htmlData[108] = &HTML_resetLimitMinsAlarm;
    htmlData[109] = &HTML_resetLimitSecsAlarm;
#endif
    reset(); // reset the algorithm
}

void TotalLeakDetection::reset() {
    HighLeakDetection::reset();
    startDayPulseTime = millis(); // start of the 24hrs
}

void TotalLeakDetection::testResetLeak() {
    time_t currentTime = millis();
    time_t deltaTime   = DELTA_TIME(currentTime, startDayPulseTime);

    if (deltaTime >= config->limitResetTime && !detected) {
        pulseCount = 0;
        startDayPulseTime = currentTime;
        ALeakDetectable::applyNewConfig(); // check if new settings should be applied
    }
}

void TotalLeakDetection::testActiveLeak() {
    detected = (pulseCount >= config->limitPulseAction);

    if (detected) {
        // store the reset % when a leak was detected
        // if it has not been done yet
        if (timeOfDetection == "") {
            timeOfDetection = DateTime::getInstance()->getDateTimeStr();
            resetWhenDetected = (int)getPercentLeakDetectionReset();

            // send e-mail to the user
            #ifdef EMAIL_NOTIFICATION
                sendEmailLeakDetected();
            #endif
        }
        active = !bypass; // if the bypass is off, close the main valve
    }
    // indication LED
    digitalWrite(TOTAL_ILOCK_LED_PIN, detected);
}

#ifdef EMAIL_NOTIFICATION
    void TotalLeakDetection::sendEmailLeakDetected() {
        EmailSender::getInstance()->sendEmail(
                EmailSender::LEAK_DETECTED,
                "Total leak detected!",
                "Total leak was just detected\n\r"
                "time of detection: " + timeOfDetection + "; "
                "limit for leak to be detected: " +
                PULSE_TO_LITER(normalConfig.limitPulseAction) + "l | " +
                PULSE_TO_LITER(alarmConfig.limitPulseAction) + "l (alarm)");
    }
#endif

float TotalLeakDetection::getPercentLeakDetectionReset() const {
    if (this->active == HIGH) return 0;

    time_t currentTime = millis();
    time_t deltaTime = DELTA_TIME(currentTime, startDayPulseTime);

    float percentDetected = deltaTime / (float)config->limitResetTime;
    percentDetected *= 100;
    return min(100, percentDetected);
}

#ifdef DEBUG
    const String TotalLeakDetection::getLogID() const {
        return "[TOTAL LEAK]";
    }
#endif

#ifdef LCD_DISPLAY
    const String TotalLeakDetection::getRow(int row) const {
        switch(row) {
            case 0:
                // 1st row on the LCD
                return getType() + " water leak";
                break;
            case 1:
                // 2nd row on the LCD
                return String("BYPASS=") + String(bypass) + String(" | ") + String("ACTIVE=") + String(active);
                break;
            case 2:
                // 3rd row on the LCD
                return String("ACTIVE=") + String(getPercentLeakDetected()) + "%";
                break;
            case 3:
                // 4th row on the LCD
                return String("RESET=") + String(getPercentLeakDetectionReset()) + "%";
                break;
            break;
        }
        return "UNTITLED"; // for any other row return "UNTITLED"
    }
#endif

#ifdef WEB_SERVER
    const String TotalLeakDetection::getHTMLData(const int id) const {
        auto fce = htmlData.find(id);
        if (fce == htmlData.end())
            return HTMLDataSource::UNDEFINED_DATA;
        return fce->second(*this);
    }

    String HTML_detectedText(const TotalLeakDetection& totalLeakDetection) {
        if (totalLeakDetection.active == LOW)
            return String("Undetected");
        return String("Detected");
    }

    String HTML_bypass(const TotalLeakDetection& totalLeakDetection) {
        if (totalLeakDetection.bypass == LOW)
            return String("OFF");
        return String("ON");
    }

    String HTML_detected(const TotalLeakDetection& totalLeakDetection) {
        if (totalLeakDetection.active)
            return String("ON");
        return String("OFF");
    }

    String HTML_activePercentage(const TotalLeakDetection& totalLeakDetection) {
        return String((int)totalLeakDetection.getPercentLeakDetected());
    }

    String HTML_resetPercentage(const TotalLeakDetection& totalLeakDetection) {
        if (totalLeakDetection.detected)
            return String(totalLeakDetection.resetWhenDetected);
        return String((int)totalLeakDetection.getPercentLeakDetectionReset());
    }

    String HTML_activePercentageColor(const TotalLeakDetection& totalLeakDetection) {
        int percentage = (int)totalLeakDetection.getPercentLeakDetected();
        if (percentage < 60) return "rgba(44, 62, 80, 0.301)";
        if (percentage >= 60 && percentage < 75) return "#ffdc52";
        if (percentage >= 75 && percentage < 85) return "#fca758";
        return "#fc5858";
    }

    String HTML_detectedTime(const TotalLeakDetection& totalLeakDetection) {
        return totalLeakDetection.timeOfDetection;
    }

    String HTML_detectionLimit(const TotalLeakDetection& totalLeakDetection) {
        return String(PULSE_TO_LITER(totalLeakDetection.normalConfig.limitPulseAction));
    }

    String HTML_resetLimitDays(const TotalLeakDetection& totalLeakDetection) {
        return String(DAYS(totalLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_resetLimitHours(const TotalLeakDetection& totalLeakDetection) {
        return String(HOURS(totalLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_resetLimitMins(const TotalLeakDetection& totalLeakDetection) {
        return String(MINS(totalLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_resetLimitSecs(const TotalLeakDetection& totalLeakDetection) {
        return String(SECS(totalLeakDetection.normalConfig.limitResetTime));
    }

    String HTML_detectionLimitAlarm(const TotalLeakDetection& totalLeakDetection) {
        return String(PULSE_TO_LITER(totalLeakDetection.alarmConfig.limitPulseAction));
    }

    String HTML_resetLimitDaysAlarm(const TotalLeakDetection& totalLeakDetection) {
        return String(DAYS(totalLeakDetection.alarmConfig.limitResetTime));
    }

    String HTML_resetLimitHoursAlarm(const TotalLeakDetection& totalLeakDetection) {
        return String(HOURS(totalLeakDetection.alarmConfig.limitResetTime));
    }

    String HTML_resetLimitMinsAlarm(const TotalLeakDetection& totalLeakDetection) {
        return String(MINS(totalLeakDetection.alarmConfig.limitResetTime));
    }

    String HTML_resetLimitSecsAlarm(const TotalLeakDetection& totalLeakDetection) {
        return String(SECS(totalLeakDetection.alarmConfig.limitResetTime));
    }
#endif