#include "WebServer.h"

WebServer* WebServer::instance = NULL;

WebServer::WebServer() {
    online = 0; // the server needs to be initialized first
}

WebServer* WebServer::getInstance() {
    if (instance == NULL)
        instance = new WebServer;
    return instance;
}

int WebServer::setEthernet() {
    pinMode(SS_SD_CARD, OUTPUT);
    pinMode(SS_ETHERNET, OUTPUT);
    digitalWrite(SS_SD_CARD, HIGH);  // SD Card not active
    digitalWrite(SS_ETHERNET, HIGH); // Ethernet not active

    Ethernet.begin(mac); // set the mac address of the shield

    // shield not found 
    if (Ethernet.hardwareStatus() == EthernetNoHardware) {
        #ifdef DEBUG
            Serial.println("Ethernet shield was not found");
        #endif
        return 0;
    }

    // ethernet cabel not connected
    if (Ethernet.linkStatus() == LinkOFF) {
        #ifdef DEBUG
            Serial.println("Ethernet cable is not connected.");
        #endif
        return 0;
    }
    return 1; // the ethernet shield has been initialized successfully
}

void WebServer::setup(LeaksController *leaksController) {
    online = 0;

    // store the leaks controller so it can be
    // update simultaneously with the webpage loading
    // off the SD card
    this->leaksController = leaksController;

    // initialize the ethernet shield
    if (!setEthernet())
        return;

    // start the server
    server.begin();
    // print out the IP address
    #ifdef DEBUG
        Serial.print("server is at ");
        Serial.println(Ethernet.localIP());
    #endif

    // check if the SD card is working
    if (!SD.begin(SS_SD_CARD)) {
        #ifdef DEBUG
            Serial.println("ERROR - SD card initialization failed!");
        #endif
        return;
    }    
    #ifdef DEBUG
        Serial.println("SUCCESS - SD card initialized.");
    #endif

    // check it there's an index.html
    // file (webpage) stored on the SD card
    if (!SD.exists(INDEX_FILE_NAME)) {
        #ifdef DEBUG
            Serial.print(INDEX_FILE_NAME);
            Serial.println(" doesn't exist.");
        #endif
        return;
    }
    #ifdef DEBUG
        Serial.print(INDEX_FILE_NAME);
        Serial.println(" exists.");
    #endif

    digitalWrite(SS_SD_CARD, HIGH); // SD Card not active
    online = 1;                     // the server is up and running
}

void WebServer::addHTMLSource(HTMLDataSource* source) {
    htmlDataSources.push_back(source);
}

void WebServer::update() {
    // don't do anything if the
    // server is not active
    if (!online)
        return;

    pinMode(SS_SD_CARD, OUTPUT);
    pinMode(SS_ETHERNET, OUTPUT);
    digitalWrite(SS_SD_CARD, HIGH);  // SD Card not active
    digitalWrite(SS_ETHERNET, HIGH); // Ethernet not active

    String httpRequest      = "";
    String httpRequestStart = "";

    EthernetClient client = server.available();

    int counter = 0;
    int state = 0;

    if (client) {
        bool currentLineIsBlank = true;
        while (client.connected()) {    // check if the client is connected
            if (client.available()) {   // and if it's available
                char c = client.read(); // read one character sent from the client
                switch (state) {
                    case 0:
                        // read first n bytes which are supposed to be "GET "
                        if (counter < (int)String(HTTP_RQ_GET_START).length()) {
                            counter++;
                            httpRequestStart += c;
                        }
                        // once you've read the first n bytes (4),
                        // check if the message says "GET "
                        if (counter == (int)String(HTTP_RQ_GET_START).length()) {
                            if (httpRequestStart == HTTP_RQ_GET_START)
                                state = 2;  // "GET " has been found, move on to processing the request
                            else state = 1; // keep looking for "GET "
                        }
                        break;
                    case 1:
                        // seek the start of the actual HTTP request ("GET ")
                        httpRequestStart += c;      // add a new character
                        httpRequestStart.remove(0); // remove the first character
                        // check if the string is "GET "
                        if (httpRequestStart == HTTP_RQ_GET_START)
                            state = 2; // "GET " has been found, move on to processing the request
                        break;
                    case 2:
                        if (c == ' ')
                            state = -1;
                        else {
                            httpRequest += c;
                            // check if the HTTP request exceeded 160B (max size)
                            if ((int)httpRequestStart.length() > MAX_HTTP_RQ_LENGTH) {
                                state = -1;
                                httpRequestStart = "";
                            }
                        }
                        break;
                }
                // reply to the client
                // HTTP header (200 OK) + the content of the website
                if (c == '\n' && currentLineIsBlank) {
                    client.println("HTTP/1.1 200 OK");
                    client.println("Content-Type: text/html");
                    client.println("Connection: close");
                    client.println();

                    webFile = SD.open(INDEX_FILE_NAME); // open the index.html file
                    if (webFile) {
                        while (webFile.available()) {
                            // keep updating the water-leak detection
                            // algorithms while the HTML content is
                            // being read off the SD card
                            leaksController->update();
    
                            buff = webFile.readStringUntil('\n');       // read nother line of the HTML file
                            client.println(replaceHTMLWithData(buff));  // insert data into the line and send it to the client
                        }
                        webFile.close(); // close the file

                    } else {
                        #ifdef DEBUG
                            Serial.print("error opening ");
                            Serial.println(INDEX_FILE_NAME);
                        #endif
                    }
                    break;
                }
                if (c == '\n')
                    currentLineIsBlank = true;
                else if (c != '\r')
                    currentLineIsBlank = false;
            }
        }
        delay(1);      // delay 1ms
        client.stop(); // dissconet the client

        // analyze the HTTP request sent from the client
        processHTTPRequest(httpRequest);
    }
}

int WebServer::processHTTPRequest(String &httpRequest) {
    // print out the HTTP request
    #ifdef DEBUG
        Serial.print("HTTP Request: ");
        Serial.println(httpRequest);
        Serial.print("Size of the HTTP request: ");
        Serial.print(httpRequest.length());
        Serial.println(" B");
    #endif

    bool acceptedHttpRequest = 0; // indication if the HTTP request is valid

    // check if the HTTP request is a valid request
    // for changing the current settings of the water-leak
    // detection algorithms
    int startPos = httpRequest.indexOf(HTTP_RQ_SETTINGS);
    if (startPos != -1) {
        acceptedHttpRequest = processHTTPRequestSettings(startPos, httpRequest, leaksController->getNumberOfLeakDetections() * 4);
    }

    // check if the HTTP request is a valid request
    // for changing the current settings the e-mail
    // notifications
    #ifdef EMAIL_NOTIFICATION
        if (acceptedHttpRequest == 0) {
            startPos = httpRequest.indexOf(HTTP_RQ_NOTIFICATION);
            if (startPos != -1) {
                acceptedHttpRequest = processHTTPRequestNotification(startPos, httpRequest, EmailSender::Type::COUNT + 1);
            }
        }
    #endif

    // if the HTTP request has been accepted
    // store the settings on the SD card
    if (acceptedHttpRequest)
        saveSettings();
    return acceptedHttpRequest;
}

void WebServer::saveSettings() {
    File file;  // the config file on the SD card

    // get all the settings to be stored on the SD card
    String data = leaksController->getFormatOfSettingsToSave() + "\n" +
                  EmailSender::getInstance()->getFormatOfSettingsToSave();

    #ifdef DEBUG
        Serial.println("saving settings...");
        Serial.println("raw data:");
        Serial.println(data);
    #endif

    // first of all, store it all into a backup file
    file = SD.open(BACKUP_SETTINGS_FILE_NAME, FILE_WRITE);
    file.println(data);
    file.close();

    // then erase the actual config file
    // and store the new settings into it 
    SD.remove(SETTINGS_FILE_NAME);
    file = SD.open(SETTINGS_FILE_NAME, FILE_WRITE);
    file.println(data);
    file.close();

    // lastly remove the backup file
    SD.remove(BACKUP_SETTINGS_FILE_NAME);

    #ifdef DEBUG
        Serial.println("settings saved successfully...");
    #endif
}

#ifdef EMAIL_NOTIFICATION
int WebServer::processHTTPRequestNotification(int startPos, String &httpRequest, int numberOfValues) {
        int numberOfSemicolons = 0;         // total number of semicolons found so far
        int length = httpRequest.length();  // length of the HTTP request
        String data[numberOfValues];        // tokens (individual parameters)
        data[0] = "";                       // the first initialization of the first parameter

        for (int i = startPos + String(HTTP_RQ_NOTIFICATION).length(); i < length; i++) {
            if (httpRequest[i] == ';') {
                // check what you have after ';' has been found (end of a parameter) validate it
                // if it's empty or it contains more than one character ('1' or '0')
                // discard it (the last one is an exception since it's an e-mail address)
                if (data[numberOfSemicolons] == "" || (numberOfSemicolons != numberOfValues - 1 && data[numberOfSemicolons].length() > 1)) 
                    return 0;

                // check if all the parameters have been read
                if (++numberOfSemicolons == numberOfValues) {
                    #ifndef UNIT_TEST
                        EmailSender::getInstance()->enableNotification(EmailSender::BOOTING, data[0] == "1");
                        EmailSender::getInstance()->enableNotification(EmailSender::DAILY_OVERVIEW, data[1] == "1");
                        EmailSender::getInstance()->enableNotification(EmailSender::LEAK_DETECTED, data[2] == "1");
                        EmailSender::getInstance()->enableNotification(EmailSender::VALVE_STATE, data[3] == "1");
                        EmailSender::getInstance()->enableNotification(EmailSender::RESET, data[4] == "1");
                        EmailSender::getInstance()->enableNotification(EmailSender::ALARM, data[5] == "1");
                        EmailSender::getInstance()->enableNotification(EmailSender::CHANGED_SETTINGS, data[6] == "1");
                        EmailSender::getInstance()->enableNotification(EmailSender::APPLIED_NEW_SETTING, data[7] == "1");
                        EmailSender::getInstance()->enableNotification(EmailSender::BYPASS, data[8] == "1");
                        EmailSender::getInstance()->setReceiverEmailAddress(data[9]);
                    #endif
                    return 1; // the HTTP request is valid
                }
                data[numberOfSemicolons] = ""; // initialize the next parameter
                continue;
            }
            data[numberOfSemicolons] += httpRequest[i]; // add another character to the current parameter
        }
        return 0;  // the HTTP request is NOT valid
    }
#endif

int WebServer::processHTTPRequestSettings(int startPos, String &httpRequest, int numberOfValues) {
    int numberOfSemicolons = 0;         // total number of semicolons found so far
    int length = httpRequest.length();  // length of the HTTP request
    String data[numberOfValues];        // tokens (individual parameters as strings)
    time_t values[numberOfValues];      // tokens (individual parameters as numbers)
    data[0] = "";                       // the first initialization of the first parameter

    for (int i = startPos + String(HTTP_RQ_SETTINGS).length(); i < length; i++) {
        // the list of allowed characters is only '0'-'9' and ';'
        // if an invalid character has been found discard the HTTP request
        if (!((httpRequest[i] >= '0' && httpRequest[i] <= '9') || httpRequest[i] == HTTP_RQ_SEPARATOR))
            return 0;

        // another separator of the parameters has been found
        if (httpRequest[i] == ';') {
            if (data[numberOfSemicolons] == "") //;; in a row (empty parameter)
                return 0;

            // all the parameters have been found
            if (++numberOfSemicolons == numberOfValues) {
                // check if all the parameters are positive 
                // integers fitting into the data type
                for (int j = 0; j < numberOfValues; j++) {
                    values[j] = strtoul(data[j].c_str(), NULL, 10);
                    if (values[j] == OUT_OF_RANGE_ERR || values[j] == 0)
                        return 0; // invalid number found
                }
                // do not apply any changes if it has been
                // called as a part of unit testing
                #ifndef UNIT_TEST
                    // high-water leak detection algorithm
                    leaksController->changeSettings(ALeakDetectable::High, LITER_TO_PULSE(values[0]), values[1], false);
                    leaksController->changeSettings(ALeakDetectable::High, LITER_TO_PULSE(values[2]), values[3], true);
                     // low-water leak detection algorithm
                    leaksController->changeSettings(ALeakDetectable::Low, values[4], values[5], false);
                    leaksController->changeSettings(ALeakDetectable::Low, values[6], values[7], true);
                     // total-water leak detection algorithms
                    leaksController->changeSettings(ALeakDetectable::Total, LITER_TO_PULSE(values[8]), values[9], false);
                    leaksController->changeSettings(ALeakDetectable::Total, LITER_TO_PULSE(values[10]), values[11], true);
                #endif
                return 1; // the HTTP request is valid
            }
            data[numberOfSemicolons] = "";  // initialize the next parameter
            continue;
        }
        data[numberOfSemicolons] += httpRequest[i]; // add another character to the current parameter
    }
    return 0; // the HTTP request is NOT valid
}

const String WebServer::replaceHTMLWithData(String html) const {
    int start;  // start of the unique sequence "%*120*%"
    int end;    // end of the unique sequence "%*120*%"
    int key;    // the id of the data
    String htmlData = HTMLDataSource::UNDEFINED_DATA; // "UNDEFINED"
    String sub; // the id as a string (it'll be yank out)
    bool found; // flag if the data associated with the id has been found

    // while there's is something to be replaced "%*120*%"
    while ((start = html.indexOf(DATA_PREF)) != -1) {
        // find the start and last index of 
        // the sequence - "%*120*%". Everything in between
        // has to be the id, e.g. 120
        start += 2;
        for (end = start; html[end] != DATA_POST[0]; end++)
            ;

        sub = html.substring(start, end);
        key = sub.toInt();                        // get the id of the data
        String old = DATA_PREF + sub + DATA_POST; // the entire string to be replaced

        // find the associated value for the id
        for (auto source : htmlDataSources) {
            htmlData = source->getHTMLData(key);
            found = false;
            if (htmlData != HTMLDataSource::UNDEFINED_DATA) {
                html.replace(old, htmlData);
                found = true;
                break;
            }
        }
        // if the value has not been found
        // replace it with "UNDEFINED" insted
        if (found == false)
            html.replace(old, htmlData);
    }
    return html;
}