#include "Arduino.h"
#include "vector"
#include "SPI.h"
#include "SD.h"

#include "Setup.h"
#include "Pins.h"
#include "LimitsDefinition.h"
#include "DateTime.h"
#include "EmailSender.h"
#include "DailyOverview.h"

#include "FreeMemoryMeasurement.h"

#ifdef GENERATE_PULSES
# include <PulseGenerator.h>
#endif

#include "PulseCounter.h"
#include "LeaksController.h"
#include "ALeakDetectable.h"
#include "LeakDetectionConfig.h"
#include "HighLeakDetection.h"
#include "LowLeakDetection.h"
#include "TotalLeakDetection.h"

#ifdef DEBUG
# include <ILoggable.h>
# include <IControllable.h>
# include <Logger.h>
#endif

#ifdef WEB_SERVER
# include <HTMLDataSource.h>
# include <WebServer.h>
# include <Consumption.h>
#endif

#ifdef LCD_DISPLAY
# include <LCDController.h>
# include <FreeMemoryMeasurement.h>
#endif 

#ifdef GENERATE_PULSES
    PulseGenerator pulseGenerator(PG_DELAY_HIGH, PG_DELAY_LOW, PG_PIN);
#endif

PulseCounter pulseCounter(SENSOR_PIN);     // pulse counter
std::vector<IControllable *> controllers;  // collection of all controllers 

HighLeakDetection *highLeakDetection;      // high-water leak detection algorithm
LowLeakDetection *lowLeakDetection;        // low-water leak detection algorithm
TotalLeakDetection *totalLeakDetection;    // total-water leak detection algorithm

// initializes all three algorithms with
// the default values as defined in LimitsDefinition.h
void applyDefaultSettings() {
    // high-water leak detection algorithm
    highLeakDetection = new HighLeakDetection(
            &pulseCounter,
            ALeakDetectable::High,
            HIGH_LEAK_NORMAL_DEFAULT_CONFIG,
            HIGH_LEAK_NORMAL_DEFAULT_CONFIG);

    // low-water leak detection algorithm
    lowLeakDetection = new LowLeakDetection(
            &pulseCounter,
            ALeakDetectable::Low,
            LOW_LEAK_NORMAL_DEFAULT_CONFIG,
            LOW_LEAK_NORMAL_DEFAULT_CONFIG);

    // total-water leak detection algorithm
    totalLeakDetection = new TotalLeakDetection(
            &pulseCounter,
            ALeakDetectable::Total,
            TOTAL_LEAK_NORMAL_DEFAULT_CONFIG,
            TOTAL_LEAK_NORMAL_DEFAULT_CONFIG);
}

// splits the string given as a parameter by ';'
// and returns the tokens in a vector
std::vector<String> split(String& data) {
    std::vector<String> tokens;
    String buffer = "";
    for (int i = 0; i < (int)data.length(); i++) {
        if (data[i] == ';') {
            tokens.push_back(buffer);
            buffer = "";
        }
        else buffer += data[i];
    }
    return tokens;
}

// initializes the water-leak detection
// algorithm with the data from the file
// stored on the SD card
int createHighLeakDetection(String data) {
    std::vector<String> tokens = split(data);
    
    // make sure there's exactly 4 values
    // in the file - limit,reset (normal ,alaram)
    if (tokens.size() != 4)
        return 0;

    highLeakDetection = new HighLeakDetection(
            &pulseCounter,
            ALeakDetectable::High,
            {strtoul(tokens[0].c_str(), NULL, 10), UNUSED, atoi(tokens[1].c_str()), HIGH_BYPASS_PIN},
            {strtoul(tokens[2].c_str(), NULL, 10), UNUSED, atoi(tokens[3].c_str()), HIGH_BYPASS_PIN}
    );
    return 1; // everything's okay
}

// initializes the low-leak detection
// algorithm with the data from the file
// stored on the SD card
int createLowLeakDetection(String data) {
    std::vector<String> tokens = split(data);

    // make sure there's exactly 4 values
    // in the file - limit,reset (normal ,alaram)
    if (tokens.size() != 4)
        return 0;

    lowLeakDetection = new LowLeakDetection(
            &pulseCounter,
            ALeakDetectable::Low,
            {strtoul(tokens[0].c_str(), NULL, 10), strtoul(tokens[1].c_str(), NULL, 10), UNUSED, LOW_BYPASS_PIN},
            {strtoul(tokens[2].c_str(), NULL, 10), strtoul(tokens[3].c_str(), NULL, 10), UNUSED, LOW_BYPASS_PIN}
    );
    return 1; // everything's okay
}

// initializes the total-leak detection
// algorithm with the data from the file
// stored on the SD card
int createTotalLeakDetection(String data) {
    std::vector<String> tokens = split(data);

    // make sure there's exactly 4 values
    // in the file - limit,reset (normal ,alaram)
    if (tokens.size() != 4)
        return 0;

    totalLeakDetection = new TotalLeakDetection(
            &pulseCounter,
            ALeakDetectable::Total,
            {strtoul(tokens[0].c_str(), NULL, 10), UNUSED, atoi(tokens[1].c_str()), TOTAL_BYPASS_PIN},
            {strtoul(tokens[2].c_str(), NULL, 10), UNUSED, atoi(tokens[3].c_str()), TOTAL_BYPASS_PIN}
    );
    return 1; // everything's okay
}

// initializes the e-mail notification
// settings with the data from the file
// stored on the SD card
int setSettingsNotification(String data) {
    #ifdef EMAIL_NOTIFICATION
        std::vector<String> tokens = split(data);

        // make sure there's exactly mailSender::Type::COUNT + 1
        // on the particular line in the settings file
        if (tokens.size() != EmailSender::Type::COUNT + 1)
            return 0;

        // apply the settings
        EmailSender::getInstance()->enableNotification(EmailSender::BOOTING, tokens[0] == "1");
        EmailSender::getInstance()->enableNotification(EmailSender::DAILY_OVERVIEW, tokens[1] == "1");
        EmailSender::getInstance()->enableNotification(EmailSender::LEAK_DETECTED, tokens[2] == "1");
        EmailSender::getInstance()->enableNotification(EmailSender::VALVE_STATE, tokens[3] == "1");
        EmailSender::getInstance()->enableNotification(EmailSender::RESET, tokens[4] == "1");
        EmailSender::getInstance()->enableNotification(EmailSender::ALARM, tokens[5] == "1");
        EmailSender::getInstance()->enableNotification(EmailSender::CHANGED_SETTINGS, tokens[6] == "1");
        EmailSender::getInstance()->enableNotification(EmailSender::APPLIED_NEW_SETTING, tokens[7] == "1");
        EmailSender::getInstance()->enableNotification(EmailSender::BYPASS, tokens[8] == "1");
        EmailSender::getInstance()->setReceiverEmailAddress(tokens[9]);
    #endif
    return 1; // everything's okay
}

void loadSettings() {
    #ifdef DEBUG
        Serial.println("loading settings...");
    #endif

    // if the SD card is not avaliable,
    // apply the default settings of all
    // three water leak detection algorithms
    if (!SD.begin(SS_SD_CARD)) {
        #ifdef DEBUG
            Serial.println("ERROR - SD card initialization failed!");
        #endif
        applyDefaultSettings();
        return;
    }

    pinMode(SS_SD_CARD, OUTPUT);
    pinMode(SS_ETHERNET, OUTPUT);
    digitalWrite(SS_SD_CARD, HIGH);  // SD Card not active
    digitalWrite(SS_ETHERNET, HIGH); // Ethernet not active
    
    // search for the settings file on the SD card
    File settingsFile = SD.open(SETTINGS_FILE_NAME);

    if (settingsFile) {
        // if the any of the lines of the settings file doesn't
        // have the right syntax, automatically apply the default
        // settings for all three water-leak detection algorithms
        if (!createHighLeakDetection(settingsFile.readStringUntil('\n')) ||
            !createLowLeakDetection(settingsFile.readStringUntil('\n')) ||
            !createTotalLeakDetection(settingsFile.readStringUntil('\n')) ||
            !setSettingsNotification(settingsFile.readStringUntil('\n')))
        {
            applyDefaultSettings();
        }
        settingsFile.close(); // close the config file

    } else {
        #ifdef DEBUG
            Serial.print("error opening ");
            Serial.println(SETTINGS_FILE_NAME);
        #endif
        applyDefaultSettings(); // apply the default settings
    }
}

void setup() {
    // enable the serial communication
    // for the terminal (debugging)
    Serial.begin(SERIAL_BAUD_RATE);
    while (!Serial)
        ;

    DateTime::getInstance(); //setup the datetime

    #ifdef DEBUG
        Serial.println("<[STARTING PROGRAM]>");
        Serial.println("pins configuration...");
    #endif

    CONFIG_PINS     // configure pins
    loadSettings(); // load setting off the SD card

    // initialize the logging part of the system
    #ifdef DEBUG
        Serial.println("adding instances to logging...");
        Logger::getInstance()->addEntity(highLeakDetection);
        Logger::getInstance()->addEntity(lowLeakDetection);
        Logger::getInstance()->addEntity(totalLeakDetection);
        Logger::getInstance()->addEntity(LeaksController::getInstance());
        Logger::getInstance()->addEntity(FreeMemoryMeasurement::getInstance());
        controllers.push_back(Logger::getInstance());
    #endif
    
    // initialize all controllers of the system
    #ifdef DEBUG
        Serial.println("adding instances of all the controllers...");
    #endif
    LeaksController::getInstance()->setPulseCounter(&pulseCounter);
    LeaksController::getInstance()->addLeakDetection(ALeakDetectable::High, highLeakDetection);
    LeaksController::getInstance()->addLeakDetection(ALeakDetectable::Low, lowLeakDetection);
    LeaksController::getInstance()->addLeakDetection(ALeakDetectable::Total, totalLeakDetection);
    controllers.push_back(LeaksController::getInstance());

    // initialize the LCD display (what pages will be displayed)
    #ifdef LCD_DISPLAY
        LCDController::getInstance()->addItem(highLeakDetection);
        LCDController::getInstance()->addItem(lowLeakDetection);
        LCDController::getInstance()->addItem(totalLeakDetection);
        LCDController::getInstance()->addItem(LeaksController::getInstance());
        //LCDController::getInstance()->addItem(FreeMemoryMeasurement::getInstance());
        controllers.push_back(LCDController::getInstance());
        #ifdef DEBUG
            Logger::getInstance()->addEntity(LCDController::getInstance());
        #endif
    #endif

    // initialize the webserver
    #ifdef WEB_SERVER
        #ifdef DEBUG
            Serial.println("web server initialization...");
        #endif
        
        // set HTML data sources
        WebServer::getInstance()->setup(LeaksController::getInstance());
        WebServer::getInstance()->addHTMLSource(highLeakDetection);
        WebServer::getInstance()->addHTMLSource(lowLeakDetection);
        WebServer::getInstance()->addHTMLSource(totalLeakDetection);
        WebServer::getInstance()->addHTMLSource(&pulseCounter);
        WebServer::getInstance()->addHTMLSource(LeaksController::getInstance());
        WebServer::getInstance()->addHTMLSource(FreeMemoryMeasurement::getInstance());
        WebServer::getInstance()->addHTMLSource(DateTime::getInstance());

        controllers.push_back(WebServer::getInstance());

        // monitoring of water consumption
        Consumption *dailyWaterConsumption = new Consumption(&pulseCounter, Consumption::DAY);
        Consumption *monthlyWaterConsumption = new Consumption(&pulseCounter, Consumption::MONTH);
        
        LeaksController::getInstance()->setDailyConsumptionCounter(dailyWaterConsumption);
        LeaksController::getInstance()->setMonthlyConsumptionCounter(monthlyWaterConsumption);

        WebServer::getInstance()->addHTMLSource(EmailSender::getInstance());
    #endif

    // initialize the daily overview class
    #ifdef LCD_DISPLAY
        DailyOverview *dailyOverview = new DailyOverview(highLeakDetection,        // high-water algo.
                                                         lowLeakDetection,         // low-water algo.        
                                                         totalLeakDetection,       // total-water algo. 
                                                         dailyWaterConsumption,    // daily consump.
                                                         monthlyWaterConsumption); // monthly consump.
        controllers.insert(controllers.begin(), dailyOverview);
        LCDController::getInstance()->addItem(dailyOverview);
    #endif

    #ifdef DEBUG
        Serial.println("program is now running...");
    #endif

    // send an e-mail that the device has started
    #ifdef EMAIL_NOTIFICATION
        EmailSender::getInstance()->sendEmail(
                EmailSender::BOOTING,
                "Home Water Leaks Detection",
                "The device is now running\n\r"
                "start time: " + DateTime::getInstance()->getDateTimeStr());
    #endif
}

void loop() {
// WARNING doesn't work correctly when the user requests to open the website
// because it's not included the the class LeakController that is updated
// independently while the website is being loaded. This is only for
// testing purposes anyway.
#ifdef GENERATE_PULSES
    pulseGenerator.update();
#endif

    // update all the constrollers of the system
    for (const auto controller : controllers)
        controller->update();
}